"""
Mutability Cython Implementation
"""
##########################################################################
#
#   Sage: Open Source Mathematical Software
#
#       Copyright (C) 2006 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
##########################################################################

from sage.misc.decorators import sage_wraps

cdef class Mutability:
    r"""
    Class to mix in mutability feature.

    EXAMPLES::

        sage: class A(SageObject, Mutability):
        ....:     def __init__(self, val):
        ....:         self._val = val
        ....:     def change(self, val):
        ....:         self._require_mutable()
        ....:         self._val = val
        ....:     def __hash__(self):
        ....:         self._require_immutable()
        ....:         return hash(self._val)
        sage: a = A(4)
        sage: a._val
        4
        sage: a.change(6); a._val
        6
        sage: hash(a)
        Traceback (most recent call last):
        ...
        ValueError: object is mutable; please make it immutable first
        sage: a.set_immutable()
        sage: a.change(4)
        Traceback (most recent call last):
        ...
        ValueError: object is immutable; please change a copy instead
        sage: hash(a)
        6

    """

    def __init__(self, is_immutable=False):
        r"""
        TESTS::

            sage: class A(SageObject, Mutability):
            ....:     def __init__(self, val):
            ....:         self._val = val
            ....:     def change(self, val):
            ....:         self._require_mutable()
            ....:         self._val = val
            ....:     def __hash__(self):
            ....:         self._require_immutable()
            ....:         return hash(self._val)
            sage: a = A(4)
            sage: TestSuite(a).run(skip ="_test_pickling")

        """
        self._is_immutable = is_immutable

    cpdef _require_mutable(self):
        r"""
        Whenever mutability is required, this method can be called.
    
        EXAMPLES::
    
            sage: class A(SageObject, Mutability):
            ....:     def __init__(self, val):
            ....:         self._val = val
            ....:     def change(self, val):
            ....:         self._require_mutable()
            ....:         self._val = val
            ....:     def __hash__(self):
            ....:         self._require_immutable()
            ....:         return hash(self._val)
            sage: a = A(4)
            sage: a.set_immutable()
            sage: a.change(6)
            Traceback (most recent call last):
            ...
            ValueError: object is immutable; please change a copy instead
    
        """
        if self._is_immutable:
            raise ValueError("object is immutable; please change a copy instead")

    cpdef _require_immutable(self):
        r"""
        Whenever immutability is required, this method can be called.
        
        EXAMPLES::
        
            sage: class A(SageObject, Mutability):
            ....:     def __init__(self, val):
            ....:         self._val = val
            ....:     def change(self, val):
            ....:         self._require_mutable()
            ....:         self._val = val
            ....:     def __hash__(self):
            ....:         self._require_immutable()
            ....:         return hash(self._val)
            sage: a = A(4)
            sage: hash(a)
            Traceback (most recent call last):
            ...
            ValueError: object is mutable; please make it immutable first
        
        """
        if not self._is_immutable:
            raise ValueError("object is mutable; please make it immutable first")

    def set_immutable(self):
        """
        Make this object immutable, so it can never again be changed.

        EXAMPLES::

            sage: v = Sequence([1,2,3,4/5])
            sage: v[0] = 5
            sage: v
            [5, 2, 3, 4/5]
            sage: v.set_immutable()
            sage: v[3] = 7
            Traceback (most recent call last):
            ...
            ValueError: object is immutable; please change a copy instead.
        """
        self._is_immutable = 1

    cpdef bint is_immutable(self):
        """
        Return ``True`` if this object is immutable (cannot be changed)
        and ``False`` if it is not.

        To make this object immutable use self.set_immutable().

        EXAMPLES::

            sage: v = Sequence([1,2,3,4/5])
            sage: v[0] = 5
            sage: v
            [5, 2, 3, 4/5]
            sage: v.is_immutable()
            False
            sage: v.set_immutable()
            sage: v.is_immutable()
            True
        """
        return self._is_immutable

    cpdef bint is_mutable(self):
        """
        Return ``True`` if this object is mutable (can be changed)
        and ``False`` if it is not.
        
        To make this object immutable use ``self.set_immutable()``.
        
        EXAMPLES::
        
            sage: v = Sequence([1,2,3,4/5])
            sage: v[0] = 5
            sage: v
            [5, 2, 3, 4/5]
            sage: v.is_mutable()
            True
            sage: v.set_immutable()
            sage: v.is_mutable()
            False
        """
        return not self._is_immutable

    def __getstate__(self):
        r"""
        Get the current state of ``self`` including the mutability status.
        
        TESTS::
        
            sage: class A(SageObject, Mutability):
            ....:     def __init__(self, val):
            ....:         self._val = val
            ....:     def change(self, val):
            ....:         self._require_mutable()
            ....:         self._val = val
            ....:     def __hash__(self):
            ....:         self._require_immutable()
            ....:         return hash(self._val)
            sage: a = A(4)
            sage: a.__dict__
            {'_val': 4}
            sage: a.__getstate__()
            {'_is_immutable': False, '_val': 4}
            sage: a.__reduce__()  # indirect doctest
            (<function _reconstructor at ...>,
             (<class '__main__.A'>,
              <class 'sage.structure.sage_object.SageObject'>,
              <sage.structure.sage_object.SageObject object at ...>),
             {'_is_immutable': False, '_val': 4})
            
        """
        state = getattr(self, '__dict__', {})
        state['_is_immutable'] = self._is_immutable
        return state

    def __setstate__(self, state):
        r"""
        Set the state of ``self`` from the dictionary ``state`` including the
        mutability status.

        TESTS::

            sage: class A(SageObject, Mutability):
            ....:     def __init__(self, val):
            ....:         self._val = val
            ....:     def change(self, val):
            ....:         self._require_mutable()
            ....:         self._val = val
            ....:     def __hash__(self):
            ....:         self._require_immutable()
            ....:         return hash(self._val)
            sage: a = A(4)
            sage: a.is_immutable()
            False
            sage: d = a.__getstate__(); d
            {'_is_immutable': False, '_val': 4}
            sage: d['_is_immutable'] = True
            sage: a.__setstate__(d)
            sage: a.is_immutable()
            True
            sage: a.__getstate__()
            {'_is_immutable': True, '_val': 4}

        """
        if hasattr(self, '__dict__'):
            self.__dict__ = state
        self._is_immutable = state['_is_immutable']

##########################################################################
## Method decorators for mutating methods resp. methods that assume immutability

def require_mutable(f):
    """
    A decorator that requires mutability for a method to be called.

    .. NOTE::

        Objects whose methods use this decorator should have an attribute
        ``_is_immutable``. Otherwise, the object is assumed to be mutable.

    EXAMPLES::

        sage: from sage.structure.mutability import require_mutable, require_immutable
        sage: class A(object):
        ....:     def __init__(self, val):
        ....:         self._m = val
        ....:     @require_mutable
        ....:     def change(self, new_val):
        ....:         'change self'
        ....:         self._m = new_val
        ....:     @require_immutable
        ....:     def __hash__(self):
        ....:         'implement hash'
        ....:         return hash(self._m)
        sage: a = A(5)
        sage: a.change(6)
        sage: hash(a)
        Traceback (most recent call last):
        ...
        ValueError: <class '__main__.A'> instance is mutable, <function ...__hash__ at ...> must not be called
        sage: a._is_immutable = True
        sage: hash(a)
        6
        sage: a.change(7)   # indirect doctest
        Traceback (most recent call last):
        ...
        ValueError: <class '__main__.A'> instance is immutable, <function ...change at ...> must not be called
        sage: from sage.misc.sageinspect import sage_getdoc
        sage: print(sage_getdoc(a.change))
        change self

    AUTHORS:

    - Simon King <simon.king@uni-jena.de>
    """
    @sage_wraps(f)
    def new_f(self, *args, **kwds):
        if getattr(self, '_is_immutable', False):
            raise ValueError("{} instance is immutable, {} must not be called".format(type(self), repr(f)))
        return f(self, *args, **kwds)
    return new_f


def require_immutable(f):
    """
    A decorator that requires immutability for a method to be called.

    .. NOTE::

        Objects whose methods use this decorator should have an attribute
        ``_is_immutable``. Otherwise, the object is assumed to be mutable.

    EXAMPLES::

        sage: from sage.structure.mutability import require_mutable, require_immutable
        sage: class A(object):
        ....:  def __init__(self, val):
        ....:      self._m = val
        ....:  @require_mutable
        ....:  def change(self, new_val):
        ....:      'change self'
        ....:      self._m = new_val
        ....:  @require_immutable
        ....:  def __hash__(self):
        ....:      'implement hash'
        ....:      return hash(self._m)
        sage: a = A(5)
        sage: a.change(6)
        sage: hash(a)   # indirect doctest
        Traceback (most recent call last):
        ...
        ValueError: <class '__main__.A'> instance is mutable, <function ...__hash__ at ...> must not be called
        sage: a._is_immutable = True
        sage: hash(a)
        6
        sage: a.change(7)
        Traceback (most recent call last):
        ...
        ValueError: <class '__main__.A'> instance is immutable, <function ...change at ...> must not be called
        sage: from sage.misc.sageinspect import sage_getdoc
        sage: print(sage_getdoc(a.__hash__))
        implement hash

    AUTHORS:

    - Simon King <simon.king@uni-jena.de>
    """
    @sage_wraps(f)
    def new_f(self, *args, **kwds):
        if not getattr(self,'_is_immutable',False):
            raise ValueError("{} instance is mutable, {} must not be called".format(type(self), repr(f)))
        return f(self, *args, **kwds)
    return new_f



r"""
Covariant Functorial Constructions

A *functorial construction* is a collection of functors
`(F_{Cat})_{Cat}` (indexed by a collection of categories) which associate
to a sequence of parents `(A, B, ...)` in a category `Cat` a parent
`F_{Cat}(A, B, ...)`. Typical examples of functorial
constructions are :obj:`cartesian_product` and :obj:`tensor_product`.

The category of `F_{Cat}(A, B, ...)`, which only depends on `Cat`, is
called the (functorial) construction category.

A functorial construction is *(category)-covariant* if for every
categories `Cat` and `SuperCat`, the category of `F_{Cat}(A, B, ...)`
is a subcategory of the category of `F_{SuperCat}(A, B, ...)` whenever
`Cat` is a subcategory of `SuperCat`. A functorial construction is
*(category)-regressive* if the category of `F_{Cat}(A, B, ...)` is a
subcategory of `Cat`.

The goal of this module is to provide generic support for covariant
functorial constructions. In particular, given some parents `A`, `B`,
..., in respective categories `Cat_A`, `Cat_B`, ..., it provides tools
for calculating the best known category for the parent
`F(A,B,...)`. For examples, knowing that Cartesian products of
semigroups (resp. monoids, groups) have a semigroup (resp. monoid,
group) structure, and given a group `B` and two monoids `A` and `C` it
can calculate that `A \times B \times C` is naturally endowed with a
monoid structure.

See :class:`CovariantFunctorialConstruction`,
:class:`CovariantConstructionCategory` and
:class:`RegressiveCovariantConstructionCategory` for more details.

AUTHORS:

 - Nicolas M. Thiery (2010): initial revision
"""
#*****************************************************************************
#  Copyright (C) 2010 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************
from sage.misc.cachefunc import cached_function, cached_method
from sage.misc.lazy_attribute import lazy_class_attribute
from sage.misc.lazy_import import LazyImport
from sage.categories.category import Category
from sage.structure.sage_object import SageObject
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.dynamic_class import DynamicMetaclass

class CovariantFunctorialConstruction(UniqueRepresentation, SageObject):
    r"""
    An abstract class for construction functors `F` (eg `F` = Cartesian
    product, tensor product, `\QQ`-algebra, ...) such that:

     - Each category `Cat` (eg `Cat=` ``Groups()``) can provide a category
       `F_{Cat}` for parents constructed via this functor (e.g.
       `F_{Cat} =` ``CartesianProductsOf(Groups())``).

     - For every category `Cat`, `F_{Cat}` is a subcategory of
       `F_{SuperCat}` for every super category `SuperCat` of
       `Cat` (the functorial construction is (category)-covariant).

     - For parents `A`, `B`, ..., respectively in the categories
       `Cat_A`, `Cat_B`, ..., the category of `F(A,B,...)` is
       `F_{Cat}` where `Cat` is the meet of the categories `Cat_A`,
       `Cat_B`, ...,.

    This covers two slightly different use cases:

     - In the first use case, one uses directly the construction
       functor to create new parents::

         sage: tensor()  # todo: not implemented (add an example)

       or even new elements, which indirectly constructs the
       corresponding parent::

         sage: tensor(...) # todo: not implemented

     - In the second use case, one implements a parent, and then put
       it in the category `F_{Cat}` to specify supplementary
       mathematical information about that parent.

    The main purpose of this class is to handle automatically the
    trivial part of the category hierarchy. For example,
    ``CartesianProductsOf(Groups())`` is set automatically as a
    subcategory of ``CartesianProductsOf(Monoids())``.

    In practice, each subclass of this class should provide the
    following attributes:

     - ``_functor_category`` - a string which should match the name of
       the nested category class to be used in each category to
       specify information and generic operations for elements of this
       category.

     - ``_functor_name`` - an string which specifies the name of the
       functor, and also (when relevant) of the method on parents and
       elements used for calling the construction.

    TODO: What syntax do we want for `F_{Cat}`? For example, for the
    tensor product construction, which one do we want among (see
    chat on IRC, on 07/12/2009):

     - ``tensor(Cat)``
     - ``tensor((Cat, Cat))``
     - ``tensor.of((Cat, Cat))``
     - ``tensor.category_from_categories((Cat, Cat, Cat))``
     - ``Cat.TensorProducts()``

    The syntax ``Cat.TensorProducts()`` does not supports well multivariate
    constructions like ``tensor.of([Algebras(), HopfAlgebras(), ...])``.
    Also it forces every category to be (somehow) aware of all the
    tensorial construction that could apply to it, even those which
    are only induced from super categories.

    Note: for each functorial construction, there probably is one (or several)
    largest categories on which it applies. For example, the
    :func:`~sage.categories.cartesian_product.CartesianProducts` construction makes
    only sense for concrete categories, that is subcategories of
    ``Sets()``. Maybe we want to model this one way or the other.
    """

    def category_from_parents(self, parents):
        """
        Return the category of `F(A,B,...)` for `A,B,...` parents.

        INPUT:

         - self: a functor F
         - parents: a list (or iterable) of parents.

        EXAMPLES::

            sage: E = CombinatorialFreeModule(QQ, ["a", "b", "c"])
            sage: tensor.category_from_parents((E, E, E))
            Category of tensor products of finite dimensional vector spaces with basis over Rational Field
        """
        from sage.structure.parent import Parent
        assert(all(isinstance(parent, Parent) for parent in parents))
        # Should we pass a set of categories to reduce the cache size?
        # But then this would impose that, for any constructor, the
        # category of the result does not depend on the order/repetition
        # of the categories of the parents
        return self.category_from_categories(tuple(set(parent.category() for parent in parents)))

    @cached_method
    def category_from_categories(self, categories):
        """
        Return the category of `F(A,B,...)` for `A,B,...` parents in
        the given categories.

        INPUT:

         - ``self``: a functor `F`
         - ``categories``: a non empty tuple of categories

        EXAMPLES::

            sage: Cat1 = Rings()
            sage: Cat2 = Groups()
            sage: cartesian_product.category_from_categories((Cat1, Cat1, Cat1))
            Join of Category of rings and ...
                and Category of Cartesian products of monoids
                and Category of Cartesian products of commutative additive groups

            sage: cartesian_product.category_from_categories((Cat1, Cat2))
            Category of Cartesian products of monoids
        """
        assert(len(categories) > 0)
        return self.category_from_category(Category.meet(categories))

    def category_from_category(self, category):
        """
        Return the category of `F(A,B,...)` for `A,B,...` parents in
        ``category``.

        INPUT:

         - ``self``: a functor `F`
         - ``category``: a category

        EXAMPLES::

            sage: tensor.category_from_category(ModulesWithBasis(QQ))
            Category of tensor products of vector spaces with basis over Rational Field

        # TODO: add support for parametrized functors
        """
        return getattr(category, self._functor_category)()

    def _repr_(self):
        """
        EXAMPLES::

            sage: tensor                       # indirect doctest
            The tensor functorial construction
        """
        return "The %s functorial construction"%self._functor_name

    def __call__(self, args, **kwargs):
        """
        Functorial construction application

        INPUT:
         - ``self``: a covariant functorial construction `F`
         - ``args``: a tuple (or iterable) of parents or elements

        Returns `F(args)`

        EXAMPLES::

            sage: E = CombinatorialFreeModule(QQ, ["a", "b", "c"]); E.rename("E")
            sage: tensor((E, E, E))
            E # E # E
        """
        args = tuple(args) # a bit brute force; let's see if this becomes a bottleneck later
        assert(all( hasattr(arg, self._functor_name) for arg in args))
        assert(len(args) > 0)
        return getattr(args[0], self._functor_name)(*args[1:], **kwargs)

class FunctorialConstructionCategory(Category): # Should this be CategoryWithBase?
    """
    Abstract class for categories `F_{Cat}` obtained through a
    functorial construction
    """

    @lazy_class_attribute
    def _base_category_class(cls):
        """
        Recover the class of the base category.

        OUTPUT:

        A *tuple* whose single entry is the base category class.

        .. WARNING::

            This is only used for functorial construction categories
            that are not implemented as nested classes, and won't work
            otherwise.

        .. SEEALSO:: :meth:`__classcall__`

        EXAMPLES::

            sage: GradedModules._base_category_class
            (<class 'sage.categories.modules.Modules'>,)
            sage: GradedAlgebrasWithBasis._base_category_class
            (<class 'sage.categories.algebras_with_basis.AlgebrasWithBasis'>,)

        The reason for wrapping the base category class in a tuple is
        that, often, the base category class implements a
        :meth:`__classget__` method which would get in the way upon
        attribute access::

            sage: F = GradedAlgebrasWithBasis
            sage: F._foo = F._base_category_class[0]
            sage: F._foo
            Traceback (most recent call last):
            ...
            AssertionError: base category class for <...AlgebrasWithBasis'>
             mismatch; expected <...Algebras'>,
             got <...GradedAlgebrasWithBasis'>

        We note that because ``Algebras.WithBasis`` is not lazily imported
        on startup (see :trac:`22955`), the test fails at a different
        point in the code. However, if this import becomes lazy again, then
        the following error will be generated and can replace the above::

            sage: F._foo  # not tested
            Traceback (most recent call last):
            ...
            ValueError: could not infer axiom for the nested class
            <...AlgebrasWithBasis'> of <...GradedAlgebrasWithBasis'>

        .. TODO::

            The logic is very similar to that implemented in
            :class:`CategoryWithAxiom._base_category_class`. Find a
            way to refactor this to avoid the duplication.
        """
        module_name = cls.__module__.replace(cls._functor_category.lower() + "_","")
        import sys
        name   = cls.__name__.replace(cls._functor_category, "")
        __import__(module_name)
        module = sys.modules[module_name]
        return (module.__dict__[name],)

    @staticmethod
    def __classcall__(cls, category=None, *args):
        """
        Make ``XXXCat(**)`` a shorthand for ``Cat(**).XXX()``.

        EXAMPLES::

            sage: GradedModules(ZZ)   # indirect doctest
            Category of graded modules over Integer Ring
            sage: Modules(ZZ).Graded()
            Category of graded modules over Integer Ring
            sage: Modules.Graded(ZZ)
            Category of graded modules over Integer Ring
            sage: GradedModules(ZZ) is Modules(ZZ).Graded()
            True

        .. SEEALSO:: :meth:`_base_category_class`

        .. TODO::

            The logic is very similar to that implemented in
            :class:`CategoryWithAxiom.__classcall__`. Find a way to
            refactor this to avoid the duplication.
        """
        base_category_class = cls._base_category_class[0]
        if isinstance(category, base_category_class):
            return super(FunctorialConstructionCategory, cls).__classcall__(cls, category, *args)
        else:
            return cls.category_of(base_category_class(category, *args))

    @staticmethod
    def __classget__(cls, base_category, base_category_class):
        r"""
        Special binding for covariant constructions.

        This implements a hack allowing e.g. ``category.Subquotients``
        to recover the default ``Subquotients`` method defined in
        ``Category``, even if it has been overridden by a
        ``Subquotients`` class.

        EXAMPLES::

            sage: Sets.Subquotients
            <class 'sage.categories.sets_cat.Sets.Subquotients'>
            sage: Sets().Subquotients
            Cached version of <function ...Subquotients at ...>

        This method also initializes the attribute
        ``_base_category_class`` if not already set::

            sage: Sets.Subquotients._base_category_class
            (<class 'sage.categories.sets_cat.Sets'>,)

        It also forces the resolution of lazy imports (see :trac:`15648`)::

            sage: type(Algebras.__dict__["Graded"])
            <class 'sage.misc.lazy_import.LazyImport'>
            sage: Algebras.Graded
            <class 'sage.categories.graded_algebras.GradedAlgebras'>
            sage: type(Algebras.__dict__["Graded"])
            <class 'sage.misc.classcall_metaclass.ClasscallMetaclass'>

        .. TODO::

            The logic is very similar to that implemented in
            :class:`CategoryWithAxiom.__classget__`. Find a way to
            refactor this to avoid the duplication.
        """
        if base_category is not None:
            assert base_category.__class__ is base_category_class
            assert isinstance(base_category_class, DynamicMetaclass)
        if isinstance(base_category_class, DynamicMetaclass):
            base_category_class = base_category_class.__base__
        if "_base_category_class" not in cls.__dict__:
            cls._base_category_class = (base_category_class,)
        else:
            assert cls._base_category_class[0] is base_category_class, \
                "base category class for {} mismatch; expected {}, got {}".format(
                 cls, cls._base_category_class[0], base_category_class)

        # Workaround #15648: if Sets.Subquotients is a LazyImport object,
        # this forces the substitution of the object back into Sets
        # to avoid resolving the lazy import over and over
        if isinstance(base_category_class.__dict__[cls._functor_category], LazyImport):
            setattr(base_category_class, cls._functor_category, cls)
        if base_category is None:
            return cls
        return getattr(super(base_category.__class__.__base__, base_category),
                       cls._functor_category)

    @classmethod
    @cached_function
    def category_of(cls, category, *args):
        """
        Return the image category of the functor `F_{Cat}`.

        This is the main entry point for constructing the category
        `F_{Cat}` of parents `F(A,B,...)` constructed from parents
        `A,B,...` in `Cat`.

        INPUT:

         - ``cls`` -- the category class for the functorial construction `F`
         - ``category`` -- a category `Cat`
         - ``*args`` -- further arguments for the functor

        EXAMPLES::

            sage: sage.categories.tensor.TensorProductsCategory.category_of(ModulesWithBasis(QQ))
            Category of tensor products of vector spaces with basis over Rational Field

            sage: sage.categories.algebra_functor.AlgebrasCategory.category_of(FiniteMonoids(), QQ)
            Join of Category of finite dimensional algebras with basis over Rational Field
                and Category of monoid algebras over Rational Field
                and Category of finite set algebras over Rational Field
        """
        # TODO: find a better test
        # the purpose is to test whether ``category`` implements that functor
        functor_category = getattr(category.__class__, cls._functor_category)
        if isinstance(functor_category, type) and issubclass(functor_category, Category):
            return functor_category(category, *args)
        else:
            return cls.default_super_categories(category, *args)

    def __init__(self, category, *args):
        r"""
        TESTS::

            sage: from sage.categories.covariant_functorial_construction import CovariantConstructionCategory
            sage: class FooBars(CovariantConstructionCategory):
            ....:     _functor_category = "FooBars"
            ....:     _base_category_class = (Category,)
            sage: Category.FooBars = lambda self: FooBars.category_of(self)
            sage: C = FooBars(ModulesWithBasis(ZZ))
            sage: C
            Category of foo bars of modules with basis over Integer Ring
            sage: C.base_category()
            Category of modules with basis over Integer Ring
            sage: latex(C)
            \mathbf{FooBars}(\mathbf{ModulesWithBasis}_{\Bold{Z}})
            sage: import __main__; __main__.FooBars = FooBars # Fake FooBars being defined in a python module
            sage: TestSuite(C).run()
        """
        assert isinstance(category, Category)
        self._base_category = category
        self._args = args
        super(FunctorialConstructionCategory, self).__init__(*args)

    def base_category(self):
        """
        Return the base category of the category ``self``.

        For any category ``B`` = `F_{Cat}` obtained through a functorial
        construction `F`, the call ``B.base_category()`` returns the
        category `Cat`.

        EXAMPLES::

            sage: Semigroups().Quotients().base_category()
            Category of semigroups
        """
        return self._base_category

    def extra_super_categories(self):
        """
        Return the extra super categories of a construction category.

        Default implementation which returns ``[]``.

        EXAMPLES::

            sage: Sets().Subquotients().extra_super_categories()
            []
            sage: Semigroups().Quotients().extra_super_categories()
            []
        """
        return []

    def super_categories(self):
        """
        Return the super categories of a construction category.

        EXAMPLES::

            sage: Sets().Subquotients().super_categories()
            [Category of sets]
            sage: Semigroups().Quotients().super_categories()
            [Category of subquotients of semigroups, Category of quotients of sets]
        """
        return Category.join([self.__class__.default_super_categories(self.base_category(), *self._args)] +
                             self.extra_super_categories(),
                             as_list = True)

    def _repr_object_names(self):
        """
        EXAMPLES::

            sage: Semigroups().Subquotients()  # indirect doctest
            Category of subquotients of semigroups
        """
        return "%s of %s"%(Category._repr_object_names(self), self.base_category()._repr_object_names())

    def _latex_(self):
        r"""
        EXAMPLES::

            sage: latex(Semigroups().Subquotients())   # indirect doctest
            \mathbf{Subquotients}(\mathbf{Semigroups})
            sage: latex(ModulesWithBasis(QQ).TensorProducts())
            \mathbf{TensorProducts}(\mathbf{WithBasis}_{\Bold{Q}})
            sage: latex(Semigroups().Algebras(QQ))
            \mathbf{Algebras}(\mathbf{Semigroups})
        """
        from sage.misc.latex import latex
        return "\\mathbf{%s}(%s)"%(self._short_name(), latex(self.base_category()))

class CovariantConstructionCategory(FunctorialConstructionCategory):
    """
    Abstract class for categories `F_{Cat}` obtained through a
    covariant functorial construction
    """

    @classmethod
    def default_super_categories(cls, category, *args):
        r"""
        Return the default super categories of `F_{Cat}(A,B,...)` for
        `A,B,...` parents in `Cat`.

        INPUT:

         - ``cls`` -- the category class for the functor `F`
         - ``category`` -- a category `Cat`
         - ``*args`` -- further arguments for the functor

        OUTPUT: a (join) category

        The default implementation is to return the join of the
        categories of `F(A,B,...)` for `A,B,...` in turn in each of
        the super categories of ``category``.

        This is implemented as a class method, in order to be able to
        reconstruct the functorial category associated to each of the
        super categories of ``category``.

        EXAMPLES:

        Bialgebras are both algebras and coalgebras::

            sage: Bialgebras(QQ).super_categories()
            [Category of algebras over Rational Field, Category of coalgebras over Rational Field]

        Hence tensor products of bialgebras are tensor products of
        algebras and tensor products of coalgebras::

            sage: Bialgebras(QQ).TensorProducts().super_categories()
            [Category of tensor products of algebras over Rational Field,
             Category of tensor products of coalgebras over Rational Field]

        Here is how :meth:`default_super_categories` was called internally::

            sage: sage.categories.tensor.TensorProductsCategory.default_super_categories(Bialgebras(QQ))
            Join of Category of tensor products of algebras over Rational Field and Category of tensor products of coalgebras over Rational Field

        We now show a similar example, with the ``Algebra`` functor
        which takes a parameter `\QQ`::

            sage: FiniteMonoids().super_categories()
            [Category of monoids, Category of finite semigroups]
            sage: sorted(FiniteMonoids().Algebras(QQ).super_categories(), key=str)
            [Category of finite dimensional algebras with basis over Rational Field,
             Category of finite set algebras over Rational Field,
             Category of monoid algebras over Rational Field]

        Note that neither the category of *finite* semigroup algebras
        nor that of monoid algebras appear in the result; this is
        because there is currently nothing specific implemented about them.

        Here is how :meth:`default_super_categories` was called internally::

            sage: sage.categories.algebra_functor.AlgebrasCategory.default_super_categories(FiniteMonoids(), QQ)
            Join of Category of finite dimensional algebras with basis over Rational Field
                and Category of monoid algebras over Rational Field
                and Category of finite set algebras over Rational Field
        """
        return Category.join([getattr(cat, cls._functor_category)(*args)
                              for cat in category._super_categories
                              if hasattr(cat, cls._functor_category)])

    def is_construction_defined_by_base(self):
        r"""
        Return whether the construction is defined by the base of ``self``.

        EXAMPLES:

        The graded functorial construction is defined by the modules
        category. Hence this method returns ``True`` for graded
        modules and ``False`` for other graded xxx categories::

            sage: Modules(ZZ).Graded().is_construction_defined_by_base()
            True
            sage: Algebras(QQ).Graded().is_construction_defined_by_base()
            False
            sage: Modules(ZZ).WithBasis().Graded().is_construction_defined_by_base()
            False

        This is implemented as follows: given the base category `A`
        and the construction `F` of ``self``, that is ``self=A.F()``,
        check whether no super category of `A` has `F` defined.

        .. NOTE::

            Recall that, when `A` does not implement the construction
            ``F``, a join category is returned. Therefore, in such
            cases, this method is not available::

                sage: Bialgebras(QQ).Graded().is_construction_defined_by_base()
                Traceback (most recent call last):
                ...
                AttributeError: 'JoinCategory_with_category' object has no attribute 'is_construction_defined_by_base'
        """
        base = self.base_category()
        f = self._functor_category
        return not any(hasattr(C, f) for C in base.super_categories())

    def additional_structure(self):
        r"""
        Return the additional structure defined by ``self``.

        By default, a functorial construction category ``A.F()``
        defines additional structure if and only if `A` is the
        category defining `F`. The rationale is that, for a
        subcategory `B` of `A`, the fact that `B.F()` morphisms shall
        preserve the `F`-specific structure is already imposed by
        `A.F()`.

        .. SEEALSO::

            - :meth:`Category.additional_structure`.
            - :meth:`is_construction_defined_by_base`.

        EXAMPLES::

            sage: Modules(ZZ).Graded().additional_structure()
            Category of graded modules over Integer Ring
            sage: Algebras(ZZ).Graded().additional_structure()

        TESTS::

            sage: Modules(ZZ).Graded().additional_structure.__module__
            'sage.categories.covariant_functorial_construction'
        """
        if self.is_construction_defined_by_base():
            return self
        else:
            return None

class RegressiveCovariantConstructionCategory(CovariantConstructionCategory):
    """
    Abstract class for categories `F_{Cat}` obtained through a
    regressive covariant functorial construction
    """

    @classmethod
    def default_super_categories(cls, category, *args):
        """
        Return the default super categories of `F_{Cat}(A,B,...)` for
        `A,B,...` parents in `Cat`.

        INPUT:

        - ``cls`` -- the category class for the functor `F`
        - ``category`` -- a category `Cat`
        - ``*args`` -- further arguments for the functor

        OUTPUT:

        A join category.

        This implements the property that an induced subcategory is a
        subcategory.

        EXAMPLES:

        A subquotient of a monoid is a monoid, and a subquotient of
        semigroup::

            sage: Monoids().Subquotients().super_categories()
            [Category of monoids, Category of subquotients of semigroups]

        TESTS::

            sage: C = Monoids().Subquotients()
            sage: C.__class__.default_super_categories(C.base_category(), *C._args)
            Category of unital subquotients of semigroups
        """
        return Category.join([category, super(RegressiveCovariantConstructionCategory, cls).default_super_categories(category, *args)])



"""
With Realizations Covariant Functorial Construction

.. SEEALSO::

    - :func:`Sets().WithRealizations <sage.categories.with_realizations.WithRealizations>`
      for an introduction to *realizations* and *with realizations*.
    - :mod:`sage.categories.covariant_functorial_construction`
      for an introduction to covariant functorial constructions.
"""
#*****************************************************************************
#  Copyright (C) 2010-2012 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.categories.category import Category
from sage.categories.covariant_functorial_construction import RegressiveCovariantConstructionCategory


def WithRealizations(self):
    r"""
    Return the category of parents in ``self`` endowed with multiple realizations.

    INPUT:

    - ``self`` -- a category

    .. SEEALSO::

        - The documentation and code
          (:mod:`sage.categories.examples.with_realizations`) of
          ``Sets().WithRealizations().example()`` for more on how to use and
          implement a parent with several realizations.

        - Various use cases:

          - :class:`SymmetricFunctions`
          - :class:`QuasiSymmetricFunctions`
          - :class:`NonCommutativeSymmetricFunctions`
          - :class:`SymmetricFunctionsNonCommutingVariables`
          - :class:`DescentAlgebra`
          - :class:`algebras.Moebius`
          - :class:`IwahoriHeckeAlgebra`
          - :class:`ExtendedAffineWeylGroup`

        - The `Implementing Algebraic Structures
          <../../../../../thematic_tutorials/tutorial-implementing-algebraic-structures>`_
          thematic tutorial.

        - :mod:`sage.categories.realizations`

    .. NOTE:: this *function* is actually inserted as a *method* in the class
       :class:`~sage.categories.category.Category` (see
       :meth:`~sage.categories.category.Category.WithRealizations`). It is defined
       here for code locality reasons.

    EXAMPLES::

        sage: Sets().WithRealizations()
        Category of sets with realizations

    .. RUBRIC:: Parent with realizations

    Let us now explain the concept of realizations. A *parent with
    realizations* is a facade parent (see :class:`Sets.Facade`)
    admitting multiple concrete realizations where its elements are
    represented. Consider for example an algebra `A` which admits
    several natural bases::

        sage: A = Sets().WithRealizations().example(); A
        The subset algebra of {1, 2, 3} over Rational Field

    For each such basis `B` one implements a parent `P_B` which
    realizes `A` with its elements represented by expanding them on
    the basis `B`::

        sage: A.F()
        The subset algebra of {1, 2, 3} over Rational Field in the Fundamental basis
        sage: A.Out()
        The subset algebra of {1, 2, 3} over Rational Field in the Out basis
        sage: A.In()
        The subset algebra of {1, 2, 3} over Rational Field in the In basis

        sage: A.an_element()
        F[{}] + 2*F[{1}] + 3*F[{2}] + F[{1, 2}]

    If `B` and `B'` are two bases, then the change of basis from `B`
    to `B'` is implemented by a canonical coercion between `P_B` and
    `P_{B'}`::

        sage: F = A.F(); In = A.In(); Out = A.Out()
        sage: i = In.an_element(); i
        In[{}] + 2*In[{1}] + 3*In[{2}] + In[{1, 2}]
        sage: F(i)
        7*F[{}] + 3*F[{1}] + 4*F[{2}] + F[{1, 2}]
        sage: F.coerce_map_from(Out)
        Generic morphism:
          From: The subset algebra of {1, 2, 3} over Rational Field in the Out basis
          To:   The subset algebra of {1, 2, 3} over Rational Field in the Fundamental basis

    allowing for mixed arithmetic::

        sage: (1 + Out.from_set(1)) * In.from_set(2,3)
        Out[{}] + 2*Out[{1}] + 2*Out[{2}] + 2*Out[{3}] + 2*Out[{1, 2}] + 2*Out[{1, 3}] + 4*Out[{2, 3}] + 4*Out[{1, 2, 3}]

    In our example, there are three realizations::

        sage: A.realizations()
        [The subset algebra of {1, 2, 3} over Rational Field in the Fundamental basis,
         The subset algebra of {1, 2, 3} over Rational Field in the In basis,
         The subset algebra of {1, 2, 3} over Rational Field in the Out basis]

    Instead of manually defining the shorthands ``F``, ``In``, and
    ``Out``, as above one can just do::

        sage: A.inject_shorthands()
        Defining F as shorthand for The subset algebra of {1, 2, 3} over Rational Field in the Fundamental basis
        Defining In as shorthand for The subset algebra of {1, 2, 3} over Rational Field in the In basis
        Defining Out as shorthand for The subset algebra of {1, 2, 3} over Rational Field in the Out basis

    .. RUBRIC:: Rationale

    Besides some goodies described below, the role of `A` is threefold:

    - To provide, as illustrated above, a single entry point for the
      algebra as a whole: documentation, access to its properties and
      different realizations, etc.

    - To provide a natural location for the initialization of the
      bases and the coercions between, and other methods that are
      common to all bases.

    - To let other objects refer to `A` while allowing elements to be
      represented in any of the realizations.

    We now illustrate this second point by defining the polynomial
    ring with coefficients in `A`::

        sage: P = A['x']; P
        Univariate Polynomial Ring in x over The subset algebra of {1, 2, 3} over Rational Field
        sage: x = P.gen()

    In the following examples, the coefficients turn out to be all
    represented in the `F` basis::

        sage: P.one()
        F[{}]
        sage: (P.an_element() + 1)^2
        F[{}]*x^2 + 2*F[{}]*x + F[{}]

    However we can create a polynomial with mixed coefficients, and
    compute with it::

        sage: p = P([1, In[{1}], Out[{2}] ]); p
        Out[{2}]*x^2 + In[{1}]*x + F[{}]
        sage: p^2
        Out[{2}]*x^4
        + (-8*In[{}] + 4*In[{1}] + 8*In[{2}] + 4*In[{3}] - 4*In[{1, 2}] - 2*In[{1, 3}] - 4*In[{2, 3}] + 2*In[{1, 2, 3}])*x^3
        + (F[{}] + 3*F[{1}] + 2*F[{2}] - 2*F[{1, 2}] - 2*F[{2, 3}] + 2*F[{1, 2, 3}])*x^2
        + (2*F[{}] + 2*F[{1}])*x
        + F[{}]

    Note how each coefficient involves a single basis which need not
    be that of the other coefficients. Which basis is used depends on
    how coercion happened during mixed arithmetic and needs not be
    deterministic.

    One can easily coerce all coefficient to a given basis with::

        sage: p.map_coefficients(In)
        (-4*In[{}] + 2*In[{1}] + 4*In[{2}] + 2*In[{3}] - 2*In[{1, 2}] - In[{1, 3}] - 2*In[{2, 3}] + In[{1, 2, 3}])*x^2 + In[{1}]*x + In[{}]

    Alas, the natural notation for constructing such polynomials does
    not yet work::

        sage: In[{1}] * x
        Traceback (most recent call last):
        ...
        TypeError: unsupported operand parent(s) for *: 'The subset algebra of {1, 2, 3} over Rational Field in the In basis' and 'Univariate Polynomial Ring in x over The subset algebra of {1, 2, 3} over Rational Field'

    .. RUBRIC:: The category of realizations of `A`

    The set of all realizations of `A`, together with the coercion morphisms
    is a category (whose class inherits from
    :class:`~sage.categories.realizations.Category_realization_of_parent`)::

        sage: A.Realizations()
        Category of realizations of The subset algebra of {1, 2, 3} over Rational Field

    The various parent realizing `A` belong to this category::

        sage: A.F() in A.Realizations()
        True

    `A` itself is in the category of algebras with realizations::

        sage: A in Algebras(QQ).WithRealizations()
        True

    The (mostly technical) ``WithRealizations`` categories are the
    analogs of the ``*WithSeveralBases`` categories in
    MuPAD-Combinat. They provide support tools for handling the
    different realizations and the morphisms between them.

    Typically, ``VectorSpaces(QQ).FiniteDimensional().WithRealizations()``
    will eventually be in charge, whenever a coercion `\phi: A\mapsto B` is
    registered, to register `\phi^{-1}` as coercion `B \mapsto A`
    if there is none defined yet. To achieve this,
    ``FiniteDimensionalVectorSpaces`` would provide a nested class
    ``WithRealizations`` implementing the appropriate logic.

    ``WithRealizations`` is a :mod:`regressive covariant functorial
    construction <sage.categories.covariant_functorial_construction>`.
    On our example, this simply means that `A` is automatically in the
    category of rings with realizations (covariance)::

        sage: A in Rings().WithRealizations()
        True

    and in the category of algebras (regressiveness)::

        sage: A in Algebras(QQ)
        True

    .. NOTE::

        For ``C`` a category, ``C.WithRealizations()`` in fact calls
        ``sage.categories.with_realizations.WithRealizations(C)``. The
        later is responsible for building the hierarchy of the
        categories with realizations in parallel to that of their base
        categories, optimizing away those categories that do not
        provide a ``WithRealizations`` nested class. See
        :mod:`sage.categories.covariant_functorial_construction` for
        the technical details.

    .. NOTE::

        Design question: currently ``WithRealizations`` is a
        regressive construction. That is ``self.WithRealizations()``
        is a subcategory of ``self`` by default::

            sage: Algebras(QQ).WithRealizations().super_categories()
            [Category of algebras over Rational Field,
             Category of monoids with realizations,
             Category of additive unital additive magmas with realizations]

        Is this always desirable? For example,
        ``AlgebrasWithBasis(QQ).WithRealizations()`` should certainly
        be a subcategory of ``Algebras(QQ)``, but not of
        ``AlgebrasWithBasis(QQ)``. This is because
        ``AlgebrasWithBasis(QQ)`` is specifying something about the
        concrete realization.

    TESTS::

        sage: Semigroups().WithRealizations()
        Join of Category of semigroups and Category of sets with realizations
        sage: C = GradedHopfAlgebrasWithBasis(QQ).WithRealizations(); C
        Category of graded hopf algebras with basis over Rational Field with realizations
        sage: C.super_categories()
        [Join of Category of hopf algebras over Rational Field
             and Category of graded algebras over Rational Field
             and Category of graded coalgebras over Rational Field]
        sage: TestSuite(Semigroups().WithRealizations()).run()
    """
    return WithRealizationsCategory.category_of(self)

Category.WithRealizations = WithRealizations

class WithRealizationsCategory(RegressiveCovariantConstructionCategory):
    """
    An abstract base class for all categories of parents with multiple
    realizations.

    .. SEEALSO:: :func:`Sets().WithRealizations <sage.categories.with_realizations.WithRealizations>`

    The role of this base class is to implement some technical goodies, such
    as the name for that category.
    """

    _functor_category = "WithRealizations"

    def _repr_(self):
        """
        String representation.

        EXAMPLES::

            sage: C = GradedHopfAlgebrasWithBasis(QQ).WithRealizations(); C #indirect doctest
            Category of graded hopf algebras with basis over Rational Field with realizations
        """
        s = repr(self.base_category())
        return s+" with realizations"



# -*- coding: utf-8 -*-
"""
Attribute and method calling
"""

# ****************************************************************************
#       Copyright (C) 2008 Mike Hansen
#       Copyright (C) 2010, 2013 Nicolas M. Thiery
#       Copyright (C) 2018 Frédéric Chapoton
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

#############################################
# Operators
#############################################
class AttrCallObject(object):
    def __init__(self, name, args, kwds):
        """
        TESTS::

            sage: f = attrcall('core', 3); f
            *.core(3)
            sage: TestSuite(f).run()
        """
        self.name = name
        self.args = args
        self.kwds = kwds

    def __call__(self, x, *args):
        """
        Gets the ``self.name`` method from ``x``, calls it with
        ``self.args`` and ``args`` as positional parameters and
        ``self.kwds`` as keyword parameters, and returns the result.

        EXAMPLES::

            sage: core = attrcall('core', 3)
            sage: core(Partition([4,2]))
            [4, 2]

            sage: series = attrcall('series', x)
            sage: series(sin(x), 4)
            1*x + (-1/6)*x^3 + Order(x^4)
        """
        return getattr(x, self.name)(*(self.args + args), **self.kwds)

    def __repr__(self):
        """
        Return a string representation of this object.

        The star in the output represents the object passed into ``self``.

        EXAMPLES::

            sage: attrcall('core', 3)
            *.core(3)
            sage: attrcall('hooks', flatten=True)
            *.hooks(flatten=True)
            sage: attrcall('hooks', 3, flatten=True)
            *.hooks(3, flatten=True)
        """
        s = "*.%s(%s" % (self.name, ", ".join(map(repr, self.args)))
        if self.kwds:
            if self.args:
                s += ", "
            s += ", ".join("%s=%s" % keyvalue for keyvalue in self.kwds.items())
        s += ")"
        return s

    def __eq__(self, other):
        """
        Equality testing

        EXAMPLES::

            sage: attrcall('core', 3, flatten = True) == attrcall('core', 3, flatten = True)
            True
            sage: attrcall('core', 2) == attrcall('core', 3)
            False
            sage: attrcall('core', 2) == 1
            False
        """
        return self.__class__ == other.__class__ and self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Equality testing

        EXAMPLES::

            sage: attrcall('core', 3, flatten = True) != attrcall('core', 3, flatten = True)
            False
            sage: attrcall('core', 2) != attrcall('core', 3)
            True
            sage: attrcall('core', 2) != 1
            True
        """
        return not self == other

    def __hash__(self):
        """
        Hash value

        This method tries to ensure that, when two ``attrcall``
        objects are equal, they have the same hash value.

        .. warning:: dicts are not hashable, so we instead hash their
        items; however the order of those items might differ. The
        proper fix would be to use a frozen dict for ``kwds``, when
        frozen dicts will be available in Python.

        EXAMPLES::

            sage: x = attrcall('core', 3, flatten = True, blah = 1)
            sage: hash(x)       # random # indirect doctest
            210434060
            sage: type(hash(x))
            <class 'int'>
            sage: y = attrcall('core', 3, blah = 1, flatten = True)
            sage: hash(y) == hash(x)
            True
            sage: y = attrcall('core', 3, flatten = True, blah = 2)
            sage: hash(y) != hash(x)
            True
            sage: hash(attrcall('core', 2)) != hash(attrcall('core', 3))
            True
            sage: hash(attrcall('core', 2)) != hash(1)
            True

        Note: a missing ``__hash__`` method here used to break the
        unique representation of parents taking ``attrcall`` objects
        as input; see :trac:`8911`.
        """
        return hash((self.args, tuple(sorted(self.kwds.items()))))


def attrcall(name, *args, **kwds):
    """
    Return a callable which takes in an object, gets the method named
    name from that object, and calls it with the specified arguments
    and keywords.

    INPUT:

    -  ``name`` - a string of the name of the method you
       want to call

    -  ``args, kwds`` - arguments and keywords to be passed
       to the method

    EXAMPLES::

        sage: f = attrcall('core', 3); f
        *.core(3)
        sage: [f(p) for p in Partitions(5)]
        [[2], [1, 1], [1, 1], [3, 1, 1], [2], [2], [1, 1]]
    """
    return AttrCallObject(name, args, kwds)


def call_method(obj, name, *args, **kwds):
    """
    Call the method ``name`` on ``obj``.

    This has to exist somewhere in Python!!!

    .. SEEALSO:: :func:`operator.methodcaller` :func:`attrcal`

    EXAMPLES::

        sage: from sage.misc.call import call_method
        sage: call_method(1, "__add__", 2)
        3
    """
    return getattr(obj, name)(*args, **kwds)

from sage.misc.persist import register_unpickle_override
register_unpickle_override("sage.misc.misc", "call_method", call_method)



#*****************************************************************************
#  Copyright (C) 2009-2010 Florent Hivert <Florent.Hivert@univ-rouen.fr>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.structure.element cimport Element


# Cython-0.17.2 disallows inline cpdef in non-final classes
# This restriction will be lifted at one point, then we can set
# some of the methods to be inline again, that is,
# revert the patch form http://trac.sagemath.org/13740

cdef class ClonableElement(Element):
    cdef bint _is_immutable
    cdef bint _needs_check
    cdef long int  _hash

    cpdef bint _require_mutable(self) except -2
    cpdef bint is_mutable(self)
    cpdef bint is_immutable(self)
    cpdef set_immutable(self)

    cpdef _set_mutable(self)

    cpdef ClonableElement clone(self, bint check=?)

cdef class ClonableArray(ClonableElement):
    cdef list _list

    cpdef list _get_list(self)
    cpdef _set_list(self, list lst)
    cpdef ClonableArray __copy__(self)
    cpdef check(self)
    cpdef object _getitem(self, int key)
    cpdef _setitem(self, int key, value)
    cpdef int index(self, key, start=*, stop=*) except -1
    cpdef int count(self, key) except -1
    cpdef long int _hash_(self) except? -1

cdef class ClonableList(ClonableArray):
    cpdef append(self, el)
    cpdef extend(self, it)
    cpdef insert(self, int index, el)
    cpdef pop(self, int index=*)
    cpdef remove(self, el)

cdef class NormalizedClonableList(ClonableList):
    cpdef normalize(self)

cdef class ClonableIntArray(ClonableElement):
    cdef int _len
    cdef int* _list

    cpdef _alloc_(self, int size)
    cpdef ClonableIntArray __copy__(self)
    cpdef check(self)
    cpdef object _getitem(self, int key)
    cpdef _setitem(self, int item, value)
    cpdef int index(self, int item) except -1
    cpdef long int _hash_(self) except? -1
    cpdef list list(self)



r"""
Finite posets

Here is some terminology used in this file:

- An *order filter* (or *upper set*) of a poset `P` is a subset `S` of `P`
  such that if `x \leq y` and `x\in S` then `y\in S`.

- An *order ideal* (or *lower set*) of a poset `P` is a subset `S` of `P`
  such that if `x \leq y` and `y\in S` then `x\in S`.
"""
#*****************************************************************************
#  Copyright (C) 2011 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.misc.abstract_method import abstract_method
from sage.categories.category_with_axiom import CategoryWithAxiom

class FinitePosets(CategoryWithAxiom):
    r"""
    The category of finite posets i.e. finite sets with a partial
    order structure.

    EXAMPLES::

        sage: FinitePosets()
        Category of finite posets
        sage: FinitePosets().super_categories()
        [Category of posets, Category of finite sets]
        sage: FinitePosets().example()
        NotImplemented

    .. SEEALSO:: :class:`~sage.categories.posets.Posets`, :func:`Poset`

    TESTS::

        sage: C = FinitePosets()
        sage: C is Posets().Finite()
        True
        sage: TestSuite(C).run()

    """

    class ParentMethods:

        ##########################################################################
        # Properties of this poset

        def is_lattice(self):
            r"""
            Return whether the poset is a lattice.

            A poset is a lattice if all pairs of elements have
            both a least upper bound ("join") and a greatest lower bound
            ("meet") in the poset.

            EXAMPLES::

                sage: P = Poset([[1, 3, 2], [4], [4, 5, 6], [6], [7], [7], [7], []])
                sage: P.is_lattice()
                True

                sage: P = Poset([[1, 2], [3], [3], []])
                sage: P.is_lattice()
                True

                sage: P = Poset({0: [2, 3], 1: [2, 3]})
                sage: P.is_lattice()
                False

                sage: P = Poset({1: [2, 3, 4], 2: [5, 6], 3: [5, 7], 4: [6, 7], 5: [8, 9],
                ....:            6: [8, 10], 7: [9, 10], 8: [11], 9: [11], 10: [11]})
                sage: P.is_lattice()
                False

            TESTS::

                sage: P = Poset()
                sage: P.is_lattice()
                True

            .. SEEALSO::

                - Weaker properties: :meth:`~sage.combinat.posets.posets.FinitePoset.is_join_semilattice`,
                  :meth:`~sage.combinat.posets.posets.FinitePoset.is_meet_semilattice`
            """
            return (self.cardinality() == 0 or
                     (self.has_bottom() and self.is_join_semilattice()))

        def is_self_dual(self):
            r"""
            Return whether the poset is *self-dual*.

            A poset is self-dual if it is isomorphic to its dual poset.

            EXAMPLES::

                sage: P = Poset({1: [3, 4], 2: [3, 4]})
                sage: P.is_self_dual()
                True

                sage: P = Poset({1: [2, 3]})
                sage: P.is_self_dual()
                False

            TESTS::

                sage: P = Poset()
                sage: P.is_self_dual()
                True

            .. SEEALSO::

                - Stronger properties: :meth:`~sage.combinat.posets.lattices.FiniteLatticePoset.is_orthocomplemented` (for lattices)
                - Other: :meth:`~sage.combinat.posets.posets.FinitePoset.dual`
            """
            # Two quick checks before full isomorphic test.
            if sorted(self._hasse_diagram.in_degree()) != sorted(self._hasse_diagram.out_degree()):
                return False
            levels_orig = [len(x) for x in self._hasse_diagram.level_sets()]
            dual_poset_hasse = self._hasse_diagram.reverse()
            levels_dual = [len(x) for x in dual_poset_hasse.level_sets()]
            if levels_orig != levels_dual:
                return False
            return self._hasse_diagram.is_isomorphic(dual_poset_hasse)

        ##########################################################################
        # Properties of morphisms

        def is_poset_isomorphism(self, f, codomain):
            r"""
            Return whether `f` is an isomorphism of posets from
            ``self`` to ``codomain``.

            INPUT:

            - ``f`` -- a function from ``self`` to ``codomain``
            - ``codomain`` -- a poset

            EXAMPLES:

            We build the poset `D` of divisors of 30, and check that
            it is isomorphic to the boolean lattice `B` of the subsets
            of `\{2,3,5\}` ordered by inclusion, via the reverse
            function `f: B \to D, b \mapsto \prod_{x\in b} x`::

                sage: D = Poset((divisors(30), attrcall("divides")))
                sage: B = Poset(([frozenset(s) for s in Subsets([2,3,5])], attrcall("issubset")))
                sage: def f(b): return D(prod(b))
                sage: B.is_poset_isomorphism(f, D)
                True

            On the other hand, `f` is not an isomorphism to the chain
            of divisors of 30, ordered by usual comparison::

                sage: P = Poset((divisors(30), operator.le))
                sage: def f(b): return P(prod(b))
                sage: B.is_poset_isomorphism(f, P)
                False

            A non surjective case::

                sage: B = Poset(([frozenset(s) for s in Subsets([2,3])], attrcall("issubset")))
                sage: def f(b): return D(prod(b))
                sage: B.is_poset_isomorphism(f, D)
                False

            A non injective case::

                sage: B = Poset(([frozenset(s) for s in Subsets([2,3,5,6])], attrcall("issubset")))
                sage: def f(b): return D(gcd(prod(b), 30))
                sage: B.is_poset_isomorphism(f, D)
                False

            .. note:: since ``D`` and ``B`` are not facade posets, ``f`` is
               responsible for the conversions between integers and subsets to
               elements of ``D`` and ``B`` and back.

            .. SEEALSO:: :meth:`FiniteLatticePosets.ParentMethods.is_lattice_morphism`
            """
            image = set(f(x) for x in self)
            if len(image) != self.cardinality():
                # Not injective
                return False
            if len(image) != codomain.cardinality():
                # Not surjective
                return False
            for x in self:
                if set(f(y) for y in self.upper_covers(x)) != set(codomain.upper_covers(f(x))):
                    return False
            return True

        def is_poset_morphism(self, f, codomain):
            r"""
            Return whether `f` is a morphism of posets from ``self``
            to ``codomain``, that is

            .. MATH::

                x\leq y \Longrightarrow f(x) \leq f(y)

            for all `x` and `y` in ``self``.

            INPUT:

            - ``f`` -- a function from ``self`` to ``codomain``
            - ``codomain`` -- a poset

            EXAMPLES:

            We build the boolean lattice of the subsets of
            `\{2,3,5,6\}` and the lattice of divisors of `30`, and
            check that the map `b \mapsto \gcd(\prod_{x\in b} x, 30)`
            is a morphism of posets::

                sage: D = Poset((divisors(30), attrcall("divides")))
                sage: B = Poset(([frozenset(s) for s in Subsets([2,3,5,6])], attrcall("issubset")))
                sage: def f(b): return D(gcd(prod(b), 30))
                sage: B.is_poset_morphism(f, D)
                True

            .. note:: since ``D`` and ``B`` are not facade posets, ``f`` is responsible
               for the conversions between integers and subsets to elements of
               ``D`` and ``B`` and back.

            `f` is also a morphism of posets to the chain of divisors
            of 30, ordered by usual comparison::

                sage: P = Poset((divisors(30), operator.le))
                sage: def f(b): return P(gcd(prod(b), 30))
                sage: B.is_poset_morphism(f, P)
                True

            FIXME: should this be ``is_order_preserving_morphism``?

            .. SEEALSO:: :meth:`is_poset_isomorphism`

            TESTS:

            Base cases::

                sage: P = posets.ChainPoset(2)
                sage: Q = posets.AntichainPoset(2)
                sage: f = lambda x: 1-x
                sage: P.is_poset_morphism(f, P)
                False
                sage: P.is_poset_morphism(f, Q)
                False
                sage: Q.is_poset_morphism(f, Q)
                True
                sage: Q.is_poset_morphism(f, P)
                True

                sage: P = Poset(); P
                Finite poset containing 0 elements
                sage: P.is_poset_morphism(f, P)
                True

            """
            for x in self:
                for y in self.upper_covers(x):
                    if not codomain.is_lequal(f(x),f(y)):
                        return False
            return True

        ##########################################################################
        # About order ideals, order filters and the like

        def order_ideal_generators(self, ideal, direction='down'):
            r"""
            Return the antichain of (minimal) generators of the order
            ideal (resp. order filter) ``ideal``.

            INPUT:

            - ``ideal`` -- an order ideal `I` (resp. order filter)
              of ``self``, as a list (or iterable); this should be
              an order ideal if ``direction`` is set to ``'down'``,
              and an order filter if ``direction`` is set to
              ``'up'``.
            - ``direction`` -- ``'up'`` or ``'down'`` (default:
              ``'down'``).

            The antichain of (minimal) generators of an order ideal
            `I` in a poset `P` is the set of all minimal elements of
            `P`. In the case of an order filter, the definition is
            similar, but with "maximal" used instead of "minimal".

            EXAMPLES:

            We build the boolean lattice of all subsets of `\{1,2,3\}`
            ordered by inclusion, and compute an order ideal there::

                sage: P = Poset((Subsets([1,2,3]), attrcall("issubset")))
                sage: I = P.order_ideal([Set([1,2]), Set([2,3]), Set([1])])
                sage: sorted(sorted(p) for p in I)
                [[], [1], [1, 2], [2], [2, 3], [3]]

            Then, we retrieve the generators of this ideal::

                sage: gen = P.order_ideal_generators(I)
                sage: sorted(sorted(p) for p in gen)
                [[1, 2], [2, 3]]

            If ``direction`` is 'up', then this instead computes
            the minimal generators for an order filter::

                sage: I = P.order_filter([Set([1,2]), Set([2,3]), Set([1])])
                sage: sorted(sorted(p) for p in I)
                [[1], [1, 2], [1, 2, 3], [1, 3], [2, 3]]
                sage: gen = P.order_ideal_generators(I, direction='up')
                sage: sorted(sorted(p) for p in gen)
                [[1], [2, 3]]

            Complexity: `O(n+m)` where `n` is the cardinality of `I`,
            and `m` the number of upper covers of elements of `I`.
            """
            if direction == 'down':
                covers = self.upper_covers
            else:
                covers = self.lower_covers
            ideal_as_set = set(ideal)
            from sage.sets.set import Set
            return Set(x for x in ideal if all(y not in ideal_as_set
                                               for y in covers(x)))

        def order_filter_generators(self, filter):
            r"""
            Generators for an order filter

            INPUT:

            - ``filter`` -- an order filter of ``self``, as a list (or iterable)

            EXAMPLES::

                sage: P = Poset((Subsets([1,2,3]), attrcall("issubset")))
                sage: I = P.order_filter([Set([1,2]), Set([2,3]), Set([1])])
                sage: sorted(sorted(p) for p in I)
                [[1], [1, 2], [1, 2, 3], [1, 3], [2, 3]]
                sage: gen = P.order_filter_generators(I)
                sage: sorted(sorted(p) for p in gen)
                [[1], [2, 3]]

            .. SEEALSO:: :meth:`order_ideal_generators`
            """
            return self.order_ideal_generators(filter, direction='up')

        def order_ideal_complement_generators(self, antichain, direction='up'):
            r"""
            Return the Panyushev complement of the antichain
            ``antichain``.

            Given an antichain `A` of a poset `P`, the Panyushev
            complement of `A` is defined to be the antichain consisting
            of the minimal elements of the order filter `B`, where `B`
            is the (set-theoretic) complement of the order ideal of
            `P` generated by `A`.

            Setting the optional keyword variable ``direction`` to
            ``'down'`` leads to the inverse Panyushev complement being
            computed instead of the Panyushev complement. The inverse
            Panyushev complement of an antichain `A` is the antichain
            whose Panyushev complement is `A`. It can be found as the
            antichain consisting of the maximal elements of the order
            ideal `C`, where `C` is the (set-theoretic) complement of
            the order filter of `P` generated by `A`.

            :meth:`panyushev_complement` is an alias for this method.

            Panyushev complementation is related (actually, isomorphic)
            to rowmotion (:meth:`rowmotion`).

            INPUT:

            - ``antichain`` -- an antichain of ``self``, as a list (or
              iterable), or, more generally, generators of an order ideal
              (resp. order filter)
            - ``direction`` -- 'up' or 'down' (default: 'up')

            OUTPUT:

            - the generating antichain of the complement order filter
              (resp. order ideal) of the order ideal (resp. order filter)
              generated by the antichain ``antichain``

            EXAMPLES::

                sage: P = Poset( ( [1,2,3], [ [1,3], [2,3] ] ) )
                sage: P.order_ideal_complement_generators([1])
                {2}
                sage: P.order_ideal_complement_generators([3])
                set()
                sage: P.order_ideal_complement_generators([1,2])
                {3}
                sage: P.order_ideal_complement_generators([1,2,3])
                set()

                sage: P.order_ideal_complement_generators([1], direction="down")
                {2}
                sage: P.order_ideal_complement_generators([3], direction="down")
                {1, 2}
                sage: P.order_ideal_complement_generators([1,2], direction="down")
                set()
                sage: P.order_ideal_complement_generators([1,2,3], direction="down")
                set()

            .. WARNING::

                This is a brute force implementation, building the
                order ideal generated by the antichain, and searching
                for order filter generators of its complement
            """
            if direction == 'up':
                I = self.order_ideal(antichain)
            else:
                I = self.order_filter(antichain)
            I_comp = set(self).difference(I)
            return set(self.order_ideal_generators(I_comp, direction = direction))

        panyushev_complement = order_ideal_complement_generators

        def rowmotion(self, order_ideal):
            r"""
            The image of the order ideal ``order_ideal`` under rowmotion
            in ``self``.

            Rowmotion on a finite poset `P` is an automorphism of the set
            `J(P)` of all order ideals of `P`. One way to define it is as
            follows: Given an order ideal `I \in J(P)`, we let `F` be the
            set-theoretic complement of `I` in `P`. Furthermore we let
            `A` be the antichain consisting of all minimal elements of
            `F`. Then, the rowmotion of `I` is defined to be the order
            ideal of `P` generated by the antichain `A` (that is, the
            order ideal consisting of each element of `P` which has some
            element of `A` above it).

            Rowmotion is related (actually, isomorphic) to Panyushev
            complementation (:meth:`panyushev_complement`).

            INPUT:

            - ``order_ideal`` -- an order ideal of ``self``, as a set

            OUTPUT:

            - the image of ``order_ideal`` under rowmotion, as a set again

            EXAMPLES::

                sage: P = Poset( {1: [2, 3], 2: [], 3: [], 4: [8], 5: [], 6: [5], 7: [1, 4], 8: []} )
                sage: I = Set({2, 6, 1, 7})
                sage: P.rowmotion(I)
                {1, 3, 4, 5, 6, 7}

                sage: P = Poset( {} )
                sage: I = Set({})
                sage: P.rowmotion(I)
                {}
            """
            result = order_ideal
            for i in reversed(self.linear_extension()):
                result = self.order_ideal_toggle(result, i)
            return result

        def birational_free_labelling(self, linear_extension=None,
                                      prefix='x', base_field=None,
                                      reduced=False, addvars=None,
                                      labels=None,
                                      min_label=None,
                                      max_label=None):
            r"""
            Return the birational free labelling of ``self``.

            Let us hold back defining this, and introduce birational
            toggles and birational rowmotion first. These notions have
            been introduced in [EP2013]_ as generalizations of the notions
            of toggles (:meth:`~sage.categories.posets.Posets.ParentMethods.order_ideal_toggle`)
            and :meth:`rowmotion <rowmotion>` on order ideals of a finite poset. They
            have been studied further in [GR2013]_.

            Let `\mathbf{K}` be a field, and `P` be a finite poset. Let
            `\widehat{P}` denote the poset obtained from `P` by adding a
            new element `1` which is greater than all existing elements
            of `P`, and a new element `0` which is smaller than all
            existing elements of `P` and `1`. Now, a `\mathbf{K}`-*labelling
            of* `P` will mean any function from `\widehat{P}` to `\mathbf{K}`.
            The image of an element `v` of `\widehat{P}` under this labelling
            will be called the *label* of this labelling at `v`. The set
            of all `\mathbf{K}`-labellings of `P` is clearly
            `\mathbf{K}^{\widehat{P}}`.

            For any `v \in P`, we now define a rational map
            `T_v : \mathbf{K}^{\widehat{P}} \dashrightarrow
            \mathbf{K}^{\widehat{P}}` as follows: For every `f \in
            \mathbf{K}^{\widehat{P}}`, the image `T_v f` should send every
            element `u \in \widehat{P}` distinct from `v` to `f(u)` (so the
            labels at all `u \neq v` don't change), while `v` is sent to

            .. MATH::

                \frac{1}{f(v)} \cdot
                \frac{\sum_{u \lessdot v} f(u)}
                {\sum_{u \gtrdot v} \frac{1}{f(u)}}

            (both sums are over all `u \in \widehat{P}` satisfying the
            respectively given conditions). Here, `\lessdot` and `\gtrdot`
            mean (respectively) "covered by" and "covers", interpreted with
            respect to the poset `\widehat{P}`. This rational map `T_v`
            is an involution and is called the *(birational)* `v`-*toggle*; see
            :meth:`birational_toggle` for its implementation.

            Now, *birational rowmotion* is defined as the composition
            `T_{v_1} \circ T_{v_2} \circ \cdots \circ T_{v_n}`, where
            `(v_1, v_2, \ldots, v_n)` is a linear extension of `P`
            (written as a linear ordering of the elements of `P`). This
            is a rational map
            `\mathbf{K}^{\widehat{P}} \dashrightarrow \mathbf{K}^{\widehat{P}}`
            which does not depend on the choice of the linear extension;
            it is denoted by `R`. See :meth:`birational_rowmotion` for
            its implementation.

            The definitions of birational toggles and birational
            rowmotion extend to the case of `\mathbf{K}` being any semifield
            rather than necessarily a field (although it becomes less
            clear what constitutes a rational map in this generality).
            The most useful case is that of the :class:`tropical semiring
            <sage.rings.semirings.tropical_semiring.TropicalSemiring>`,
            in which case birational rowmotion relates to classical
            constructions such as promotion of rectangular semistandard
            Young tableaux (page 5 of [EP2013b]_ and future work, via the
            related notion of birational *promotion*) and rowmotion on
            order ideals of the poset ([EP2013]_).

            The *birational free labelling* is a special labelling
            defined for every finite poset `P` and every linear extension
            `(v_1, v_2, \ldots, v_n)` of `P`. It is given by sending
            every element `v_i` in `P` to `x_i`, sending the element `0`
            of `\widehat{P}` to `a`, and sending the element `1` of
            `\widehat{P}` to `b`, where the ground field `\mathbf{K}` is the
            field of rational functions in `n+2` indeterminates
            `a, x_1, x_2, \ldots, x_n, b` over `\mathbb Q`.

            In Sage, a labelling `f` of a poset `P` is encoded as a
            `4`-tuple `(\mathbf{K}, d, u, v)`, where `\mathbf{K}` is the
            ground field of the labelling (i. e., its target), `d` is the
            dictionary containing the values of `f` at the elements of
            `P` (the keys being the respective elements of `P`), `u`
            is the label of `f` at `0`, and `v` is the label of `f` at
            `1`.

            .. WARNING::

                The dictionary `d` is labelled by the elements of `P`.
                If `P` is a poset with ``facade`` option set to
                ``False``, these might not be what they seem to be!
                (For instance, if
                ``P == Poset({1: [2, 3]}, facade=False)``, then the
                value of `d` at `1` has to be accessed by ``d[P(1)]``, not
                by ``d[1]``.)

            .. WARNING::

                Dictionaries are mutable. They do compare correctly,
                but are not hashable and need to be cloned to avoid
                spooky action at a distance. Be careful!

            INPUT:

            - ``linear_extension`` -- (default: the default linear
              extension of ``self``) a linear extension of ``self``
              (as a linear extension or as a list), or more generally
              a list of all elements of all elements of ``self`` each
              occurring exactly once

            - ``prefix`` -- (default: ``'x'``) the prefix to name
              the indeterminates corresponding to the elements of
              ``self`` in the labelling (so, setting it to
              ``'frog'`` will result in these indeterminates being
              called ``frog1, frog2, ..., frogn`` rather than
              ``x1, x2, ..., xn``).

            - ``base_field`` -- (default: ``QQ``) the base field to
              be used instead of `\QQ` to define the rational
              function field over; this is not going to be the base
              field of the labelling, because the latter will have
              indeterminates adjoined!

            - ``reduced`` -- (default: ``False``) if set to
              ``True``, the result will be the *reduced* birational
              free labelling, which differs from the regular one by
              having `0` and `1` both sent to `1` instead of `a` and
              `b` (the indeterminates `a` and `b` then also won't
              appear in the ground field)

            - ``addvars`` -- (default: ``''``) a string containing
              names of extra variables to be adjoined to the ground
              field (these don't have an effect on the labels)

            - ``labels`` -- (default: ``'x'``) Either a function
              that takes an element of the poset and returns a name
              for the indeterminate corresponding to that element,
              or a string containing a comma-separated list of
              indeterminates that will be assigned to elements in
              the order of ``linear_extension``. If the
              list contains more indeterminates than needed, the
              excess will be ignored. If it contains too few, then
              the needed indeterminates will be constructed from
              ``prefix``.

            - ``min_label`` -- (default: ``'a'``) a string to be
              used as the label for the element `0` of `\widehat{P}`

            - ``max_label`` -- (default: ``'b'``) a string to be
              used as the label for the element `1` of `\widehat{P}`

            OUTPUT:

            The birational free labelling of the poset ``self`` and the
            linear extension ``linear_extension``. Or, if ``reduced``
            is set to ``True``, the reduced birational free labelling.

            EXAMPLES:

            We construct the birational free labelling on a simple
            poset::

                sage: P = Poset({1: [2, 3]})
                sage: l = P.birational_free_labelling(); l
                (Fraction Field of Multivariate Polynomial Ring in a, x1, x2, x3, b over Rational Field,
                 {...},
                 a,
                 b)
                sage: sorted(l[1].items())
                [(1, x1), (2, x2), (3, x3)]

                sage: l = P.birational_free_labelling(linear_extension=[1, 3, 2]); l
                (Fraction Field of Multivariate Polynomial Ring in a, x1, x2, x3, b over Rational Field,
                 {...},
                 a,
                 b)
                sage: sorted(l[1].items())
                [(1, x1), (2, x3), (3, x2)]

                sage: l = P.birational_free_labelling(linear_extension=[1, 3, 2], reduced=True, addvars="spam, eggs"); l
                (Fraction Field of Multivariate Polynomial Ring in x1, x2, x3, spam, eggs over Rational Field,
                 {...},
                 1,
                 1)
                sage: sorted(l[1].items())
                [(1, x1), (2, x3), (3, x2)]

                sage: l = P.birational_free_labelling(linear_extension=[1, 3, 2], prefix="wut", reduced=True, addvars="spam, eggs"); l
                (Fraction Field of Multivariate Polynomial Ring in wut1, wut2, wut3, spam, eggs over Rational Field,
                 {...},
                 1,
                 1)
                sage: sorted(l[1].items())
                [(1, wut1), (2, wut3), (3, wut2)]

                sage: l = P.birational_free_labelling(linear_extension=[1, 3, 2], reduced=False, addvars="spam, eggs"); l
                (Fraction Field of Multivariate Polynomial Ring in a, x1, x2, x3, b, spam, eggs over Rational Field,
                 {...},
                 a,
                 b)
                sage: sorted(l[1].items())
                [(1, x1), (2, x3), (3, x2)]
                sage: l[1][2]
                x3

            Illustrating labelling with a function::

                sage: P = posets.ChainPoset(2).product(posets.ChainPoset(2))
                sage: l = P.birational_free_labelling(labels=lambda e : 'x_' + str(e[0]) + str(e[1]))
                sage: sorted(l[1].items())
                [((0, 0), x_00), ((0, 1), x_01), ((1, 0), x_10), ((1, 1), x_11)]
                sage: l[2]
                a

            The same, but with ``min_label`` and ``max_label`` provided::

                sage: P = posets.ChainPoset(2).product(posets.ChainPoset(2))
                sage: l = P.birational_free_labelling(labels=lambda e : 'x_' + str(e[0]) + str(e[1]), min_label="lambda", max_label="mu")
                sage: sorted(l[1].items())
                [((0, 0), x_00), ((0, 1), x_01), ((1, 0), x_10), ((1, 1), x_11)]
                sage: l[2]
                lambda
                sage: l[3]
                mu

            Illustrating labelling with a comma separated list of labels::

                sage: l = P.birational_free_labelling(labels='w,x,y,z')
                sage: sorted(l[1].items())
                [((0, 0), w), ((0, 1), x), ((1, 0), y), ((1, 1), z)]
                sage: l = P.birational_free_labelling(labels='w,x,y,z,m')
                sage: sorted(l[1].items())
                [((0, 0), w), ((0, 1), x), ((1, 0), y), ((1, 1), z)]
                sage: l = P.birational_free_labelling(labels='w')
                sage: sorted(l[1].items())
                [((0, 0), w), ((0, 1), x1), ((1, 0), x2), ((1, 1), x3)]

            Illustrating the warning about facade::

                sage: P = Poset({1: [2, 3]}, facade=False)
                sage: l = P.birational_free_labelling(linear_extension=[1, 3, 2], reduced=False, addvars="spam, eggs"); l
                (Fraction Field of Multivariate Polynomial Ring in a, x1, x2, x3, b, spam, eggs over Rational Field,
                 {...},
                 a,
                 b)
                sage: l[1][2]
                Traceback (most recent call last):
                ...
                KeyError: 2
                sage: l[1][P(2)]
                x3

            Another poset::

                sage: P = posets.SSTPoset([2,1])
                sage: lext = sorted(P)
                sage: l = P.birational_free_labelling(linear_extension=lext, addvars="ohai")
                sage: l
                (Fraction Field of Multivariate Polynomial Ring in a, x1, x2, x3, x4, x5, x6, x7, x8, b, ohai over Rational Field,
                 {...},
                 a,
                 b)
                sage: sorted(l[1].items())
                [([[1, 1], [2]], x1), ([[1, 1], [3]], x2), ([[1, 2], [2]], x3), ([[1, 2], [3]], x4),
                 ([[1, 3], [2]], x5), ([[1, 3], [3]], x6), ([[2, 2], [3]], x7), ([[2, 3], [3]], x8)]

            See :meth:`birational_rowmotion`, :meth:`birational_toggle` and
            :meth:`birational_toggles` for more substantial examples of what
            one can do with the birational free labelling.

            TESTS:

            The ``linear_extension`` keyword does not have to be given an
            actual linear extension::

                sage: P = posets.ChainPoset(2).product(posets.ChainPoset(3))
                sage: P
                Finite lattice containing 6 elements
                sage: lex = [(1,0),(0,0),(1,1),(0,1),(1,2),(0,2)]
                sage: l = P.birational_free_labelling(linear_extension=lex,
                ....:                                 prefix="u", reduced=True)
                sage: l
                (Fraction Field of Multivariate Polynomial Ring in u1, u2, u3, u4, u5, u6 over Rational Field,
                 {...},
                 1,
                 1)
                sage: sorted(l[1].items())
                [((0, 0), u2),
                 ((0, 1), u4),
                 ((0, 2), u6),
                 ((1, 0), u1),
                 ((1, 1), u3),
                 ((1, 2), u5)]

            For comparison, the standard linear extension::

                sage: l = P.birational_free_labelling(prefix="u", reduced=True); l
                (Fraction Field of Multivariate Polynomial Ring in u1, u2, u3, u4, u5, u6 over Rational Field,
                 {...},
                 1,
                 1)
                sage: sorted(l[1].items())
                [((0, 0), u1),
                 ((0, 1), u2),
                 ((0, 2), u3),
                 ((1, 0), u4),
                 ((1, 1), u5),
                 ((1, 2), u6)]

            If you want your linear extension to be tested for being a
            linear extension, just call the ``linear_extension`` method
            on the poset::

                sage: lex = [(0,0),(0,1),(1,0),(1,1),(0,2),(1,2)]
                sage: l = P.birational_free_labelling(linear_extension=P.linear_extension(lex),
                ....:                                 prefix="u", reduced=True)
                sage: l
                (Fraction Field of Multivariate Polynomial Ring in u1, u2, u3, u4, u5, u6 over Rational Field,
                 {...},
                 1,
                 1)
                sage: sorted(l[1].items())
                [((0, 0), u1),
                 ((0, 1), u2),
                 ((0, 2), u5),
                 ((1, 0), u3),
                 ((1, 1), u4),
                 ((1, 2), u6)]

            Nonstandard base field::

                sage: P = Poset({1: [3], 2: [3,4]})
                sage: lex = [1, 2, 4, 3]
                sage: l = P.birational_free_labelling(linear_extension=lex,
                ....:                                 prefix="aaa",
                ....:                                 base_field=Zmod(13))
                sage: l
                (Fraction Field of Multivariate Polynomial Ring in a, aaa1, aaa2, aaa3, aaa4, b over Ring of integers modulo 13,
                 {...},
                 a,
                 b)
                sage: l[1][4]
                aaa3

            The empty poset::

                sage: P = Poset({})
                sage: P.birational_free_labelling(reduced=False, addvars="spam, eggs")
                (Fraction Field of Multivariate Polynomial Ring in a, b, spam, eggs over Rational Field,
                 {},
                 a,
                 b)
                sage: P.birational_free_labelling(reduced=True, addvars="spam, eggs")
                (Fraction Field of Multivariate Polynomial Ring in spam, eggs over Rational Field,
                 {},
                 1,
                 1)
                sage: P.birational_free_labelling(reduced=True)
                (Multivariate Polynomial Ring in no variables over Rational Field,
                 {},
                 1,
                 1)
                sage: P.birational_free_labelling(prefix="zzz")
                (Fraction Field of Multivariate Polynomial Ring in a, b over Rational Field,
                 {},
                 a,
                 b)
                sage: P.birational_free_labelling(labels="x,y,z", min_label="spam", max_label="eggs")
                (Fraction Field of Multivariate Polynomial Ring in spam, eggs over Rational Field,
                 {},
                 spam,
                 eggs)
            """
            if base_field is None:
                from sage.rings.rational_field import QQ
                base_field = QQ
            if linear_extension is None:
                linear_extension = self.linear_extension()
            n = self.cardinality()
            label_list = []
            if labels:
                if callable(labels):
                    label_list = [labels(e) for e in linear_extension]
                else:
                    label_list = labels.split(',')
                    if len(label_list) > n:
                        label_list = label_list[:n]
                    elif len(label_list) < n:
                        label_list += [prefix + str(i) for i in range(1, n + 1 - len(label_list))]
            else:
                label_list = [prefix + str(i) for i in range(1, n + 1)]
            if not reduced:
                if min_label is None:
                    min_label = 'a'
                if max_label is None:
                    max_label = 'b'
                label_list = [min_label] + label_list + [max_label]
            if addvars:
                label_list += addvars.split(',')
            varstring = ','.join(label_list)
            varnum = len(label_list)

            from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
            PR = PolynomialRing(base_field, varstring, varnum)
            # Now, ``PR`` is the polynomial ring in `n + 2` indeterminates
            # (or more, if ``addvars`` was set; or less, if ``reduced`` is
            # ``True``) over ``base_field``.
            # The first `n + 2` of these indeterminates are named
            # ``a, x1, x2, ..., xn, b`` (if ``reduced`` is ``False``).
            # These will label the vertices of `\widehat{P}`.
            if reduced:
                xs = tuple(PR.gens()[: n])
            else:
                xs = tuple(PR.gens()[1 : n + 1])
            # So ``xs`` is the list ``[x1, x2, ..., xn]``.
            if not reduced:
                a = PR.gens()[0]
                b = PR.gens()[n + 1]
            else:
                a = PR.one()
                b = PR.one()
            # So ``a`` and ``b`` are the labels at `0` and `1`.
            FF = PR.fraction_field()
            # ``FF`` is the field of rational functions.
            dct = {self(p): xs[i] for (i, p) in enumerate(linear_extension)}
            return (FF, dct, a, b)

        def birational_toggle(self, v, labelling):
            r"""
            Return the result of applying the birational `v`-toggle `T_v`
            to the `\mathbf{K}`-labelling ``labelling`` of the poset ``self``.

            See the documentation of :meth:`birational_free_labelling`
            for a definition of this toggle and of `\mathbf{K}`-labellings as
            well as an explanation of how `\mathbf{K}`-labellings are to be
            encoded to be understood by Sage. This implementation allows
            `\mathbf{K}` to be a semifield, not just a field. The birational
            `v`-toggle is only a rational map, so an exception (most
            likely, ``ZeroDivisionError``) will be thrown if the
            denominator is zero.

            INPUT:

            - ``v`` -- an element of ``self`` (must have ``self`` as
              parent if ``self`` is a ``facade=False`` poset)

            - ``labelling`` -- a `\mathbf{K}`-labelling of ``self`` in the
              sense as defined in the documentation of
              :meth:`birational_free_labelling`

            OUTPUT:

            The `\mathbf{K}`-labelling `T_v f` of ``self``, where `f` is
            ``labelling``.

            EXAMPLES:

            Let us start with the birational free labelling of the
            "V"-poset (the three-element poset with Hasse diagram looking
            like a "V")::

                sage: V = Poset({1: [2, 3]})
                sage: s = V.birational_free_labelling(); s
                (Fraction Field of Multivariate Polynomial Ring in a, x1, x2, x3, b over Rational Field,
                 {...},
                 a,
                 b)
                sage: sorted(s[1].items())
                [(1, x1), (2, x2), (3, x3)]

            The image of `s` under the `1`-toggle `T_1` is::

                sage: s1 = V.birational_toggle(1, s); s1
                (Fraction Field of Multivariate Polynomial Ring in a, x1, x2, x3, b over Rational Field,
                 {...},
                 a,
                 b)
                sage: sorted(s1[1].items())
                [(1, a*x2*x3/(x1*x2 + x1*x3)), (2, x2), (3, x3)]

            Now let us apply the `2`-toggle `T_2` (to the old ``s``)::

                sage: s2 = V.birational_toggle(2, s); s2
                (Fraction Field of Multivariate Polynomial Ring in a, x1, x2, x3, b over Rational Field,
                 {...},
                 a,
                 b)
                sage: sorted(s2[1].items())
                [(1, x1), (2, x1*b/x2), (3, x3)]

            On the other hand, we can also apply `T_2` to the image of `s`
            under `T_1`::

                sage: s12 = V.birational_toggle(2, s1); s12
                (Fraction Field of Multivariate Polynomial Ring in a, x1, x2, x3, b over Rational Field,
                 {...},
                 a,
                 b)
                sage: sorted(s12[1].items())
                [(1, a*x2*x3/(x1*x2 + x1*x3)), (2, a*x3*b/(x1*x2 + x1*x3)), (3, x3)]

            Each toggle is an involution::

                sage: all( V.birational_toggle(i, V.birational_toggle(i, s)) == s
                ....:      for i in V )
                True

            We can also start with a less generic labelling::

                sage: t = (QQ, {1: 3, 2: 6, 3: 7}, 2, 10)
                sage: t1 = V.birational_toggle(1, t); t1
                (Rational Field, {...}, 2, 10)
                sage: sorted(t1[1].items())
                [(1, 28/13), (2, 6), (3, 7)]
                sage: t13 = V.birational_toggle(3, t1); t13
                (Rational Field, {...}, 2, 10)
                sage: sorted(t13[1].items())
                [(1, 28/13), (2, 6), (3, 40/13)]

            However, labellings have to be sufficiently generic, lest
            denominators vanish::

                sage: t = (QQ, {1: 3, 2: 5, 3: -5}, 1, 15)
                sage: t1 = V.birational_toggle(1, t)
                Traceback (most recent call last):
                ...
                ZeroDivisionError: rational division by zero

            We don't get into zero-division issues in the tropical
            semiring (unless the zero of the tropical semiring appears
            in the labelling)::

                sage: TT = TropicalSemiring(QQ)
                sage: t = (TT, {1: TT(2), 2: TT(4), 3: TT(1)}, TT(6), TT(0))
                sage: t1 = V.birational_toggle(1, t); t1
                (Tropical semiring over Rational Field, {...}, 6, 0)
                sage: sorted(t1[1].items())
                [(1, 8), (2, 4), (3, 1)]
                sage: t12 = V.birational_toggle(2, t1); t12
                (Tropical semiring over Rational Field, {...}, 6, 0)
                sage: sorted(t12[1].items())
                [(1, 8), (2, 4), (3, 1)]
                sage: t123 = V.birational_toggle(3, t12); t123
                (Tropical semiring over Rational Field, {...}, 6, 0)
                sage: sorted(t123[1].items())
                [(1, 8), (2, 4), (3, 7)]

            We turn to more interesting posets. Here is the `6`-element
            poset arising from the weak order on `S_3`::

                sage: P = posets.SymmetricGroupWeakOrderPoset(3)
                sage: sorted(list(P))
                ['123', '132', '213', '231', '312', '321']
                sage: t = (TT, {'123': TT(4), '132': TT(2), '213': TT(3), '231': TT(1), '321': TT(1), '312': TT(2)}, TT(7), TT(1))
                sage: t1 = P.birational_toggle('123', t); t1
                (Tropical semiring over Rational Field, {...}, 7, 1)
                sage: sorted(t1[1].items())
                [('123', 6), ('132', 2), ('213', 3), ('231', 1), ('312', 2), ('321', 1)]
                sage: t13 = P.birational_toggle('213', t1); t13
                (Tropical semiring over Rational Field, {...}, 7, 1)
                sage: sorted(t13[1].items())
                [('123', 6), ('132', 2), ('213', 4), ('231', 1), ('312', 2), ('321', 1)]

            Let us verify on this example some basic properties of
            toggles. First of all, again let us check that `T_v` is an
            involution for every `v`::

                sage: all( P.birational_toggle(v, P.birational_toggle(v, t)) == t
                ....:      for v in P )
                True

            Furthermore, two toggles `T_v` and `T_w` commute unless
            one of `v` or `w` covers the other::

                sage: all( P.covers(v, w) or P.covers(w, v)
                ....:      or P.birational_toggle(v, P.birational_toggle(w, t))
                ....:         == P.birational_toggle(w, P.birational_toggle(v, t))
                ....:      for v in P for w in P )
                True

            TESTS:

            Setting ``facade`` to ``False`` does not break
            ``birational_toggle``::

                sage: P = Poset({'x': ['y', 'w'], 'y': ['z'], 'w': ['z']}, facade=False)
                sage: lex = ['x', 'y', 'w', 'z']
                sage: t = P.birational_free_labelling(linear_extension=lex)
                sage: all( P.birational_toggle(v, P.birational_toggle(v, t)) == t
                ....:      for v in P )
                True
                sage: t4 = P.birational_toggle(P('z'), t); t4
                (Fraction Field of Multivariate Polynomial Ring in a, x1, x2, x3, x4, b over Rational Field,
                 {...},
                 a,
                 b)
                sage: t4[1][P('x')]
                x1
                sage: t4[1][P('y')]
                x2
                sage: t4[1][P('w')]
                x3
                sage: t4[1][P('z')]
                (x2*b + x3*b)/x4

            The one-element poset::

                sage: P = Poset({8: []})
                sage: t = P.birational_free_labelling()
                sage: t8 = P.birational_toggle(8, t); t8
                (Fraction Field of Multivariate Polynomial Ring in a, x1, b over Rational Field,
                 {...},
                 a,
                 b)
                sage: t8[1][8]
                a*b/x1
            """
            FF = labelling[0]       # base field
            a = labelling[2]        # label at `0 \in \widehat{P}`
            b = labelling[3]
            newdict = labelling[1].copy()
            # Construct the harmonic sum ``x`` of the labels at the
            # elements covering ``v``:
            uppers = self.upper_covers(v)
            if len(uppers) == 0:
                x = FF.one() / b
            else:
                x = FF.sum(FF.one() / newdict[j] for j in uppers)
                # ``FF.sum``, not ``sum``, see trac #15591.
            x = FF.one() / x
            # Construct the sum ``y`` of the labels at the elements
            # covered by ``v``:
            lowers = self.lower_covers(v)
            if len(lowers) == 0:
                y = a
            else:
                y = FF.sum(newdict[j] for j in lowers)
            # Now, transform the label at v:
            newdict[v] = x * y / newdict[v]
            return (FF, newdict, a, b)

        def birational_toggles(self, vs, labelling):
            r"""
            Return the result of applying a sequence of birational
            toggles (specified by ``vs``) to the `\mathbf{K}`-labelling
            ``labelling`` of the poset ``self``.

            See the documentation of :meth:`birational_free_labelling`
            for a definition of birational toggles and `\mathbf{K}`-labellings
            and for an explanation of how `\mathbf{K}`-labellings are to be
            encoded to be understood by Sage. This implementation allows
            `\mathbf{K}` to be a semifield, not just a field. The birational
            `v`-toggle is only a rational map, so an exception (most
            likely, ``ZeroDivisionError``) will be thrown if the
            denominator is zero.

            INPUT:

            - ``vs`` -- an iterable comprising elements of ``self``
              (which must have ``self`` as parent if ``self`` is a
              ``facade=False`` poset)

            - ``labelling`` -- a `\mathbf{K}`-labelling of ``self`` in the
              sense as defined in the documentation of
              :meth:`birational_free_labelling`

            OUTPUT:

            The `\mathbf{K}`-labelling `T_{v_n} T_{v_{n-1}} \cdots T_{v_1} f`
            of ``self``, where `f` is ``labelling`` and
            `(v_1, v_2, \ldots, v_n)` is ``vs`` (written as list).

            EXAMPLES::

                sage: P = posets.SymmetricGroupBruhatOrderPoset(3)
                sage: sorted(list(P))
                ['123', '132', '213', '231', '312', '321']
                sage: TT = TropicalSemiring(ZZ)
                sage: t = (TT, {'123': TT(4), '132': TT(2), '213': TT(3), '231': TT(1), '321': TT(1), '312': TT(2)}, TT(7), TT(1))
                sage: tA = P.birational_toggles(['123', '231', '312'], t); tA
                (Tropical semiring over Integer Ring, {...}, 7, 1)
                sage: sorted(tA[1].items())
                [('123', 6), ('132', 2), ('213', 3), ('231', 2), ('312', 1), ('321', 1)]
                sage: tAB = P.birational_toggles(['132', '213', '321'], tA); tAB
                (Tropical semiring over Integer Ring, {...}, 7, 1)
                sage: sorted(tAB[1].items())
                [('123', 6), ('132', 6), ('213', 5), ('231', 2), ('312', 1), ('321', 1)]

                sage: P = Poset({1: [2, 3], 2: [4], 3: [4]})
                sage: Qx = PolynomialRing(QQ, 'x').fraction_field()
                sage: x = Qx.gen()
                sage: t = (Qx, {1: 1, 2: x, 3: (x+1)/x, 4: x^2}, 1, 1)
                sage: t1 = P.birational_toggles((i for i in range(1, 5)), t); t1
                (Fraction Field of Univariate Polynomial Ring in x over Rational Field,
                 {...},
                 1,
                 1)
                sage: sorted(t1[1].items())
                [(1, (x^2 + x)/(x^2 + x + 1)), (2, (x^3 + x^2)/(x^2 + x + 1)), (3, x^4/(x^2 + x + 1)), (4, 1)]
                sage: t2 = P.birational_toggles(reversed(range(1, 5)), t)
                sage: sorted(t2[1].items())
                [(1, 1/x^2), (2, (x^2 + x + 1)/x^4), (3, (x^2 + x + 1)/(x^3 + x^2)), (4, (x^2 + x + 1)/x^3)]

            Facade set to ``False`` works::

                sage: P = Poset({'x': ['y', 'w'], 'y': ['z'], 'w': ['z']}, facade=False)
                sage: lex = ['x', 'y', 'w', 'z']
                sage: t = P.birational_free_labelling(linear_extension=lex)
                sage: sorted(P.birational_toggles([P('x'), P('y')], t)[1].items())
                [(x, a*x2*x3/(x1*x2 + x1*x3)), (y, a*x3*x4/(x1*x2 + x1*x3)), (w, x3), (z, x4)]
            """
            l = labelling
            for v in vs:
                l = self.birational_toggle(v, l)
            return l

        def birational_rowmotion(self, labelling):
            r"""
            Return the result of applying birational rowmotion to the
            `\mathbf{K}`-labelling ``labelling`` of the poset ``self``.

            See the documentation of :meth:`birational_free_labelling`
            for a definition of birational rowmotion and
            `\mathbf{K}`-labellings and for an explanation of how
            `\mathbf{K}`-labellings are to be encoded to be understood
            by Sage. This implementation allows `\mathbf{K}` to be a
            semifield, not just a field. Birational rowmotion is only a
            rational map, so an exception (most likely, ``ZeroDivisionError``)
            will be thrown if the denominator is zero.

            INPUT:

            - ``labelling`` -- a `\mathbf{K}`-labelling of ``self`` in the
              sense as defined in the documentation of
              :meth:`birational_free_labelling`

            OUTPUT:

            The image of the `\mathbf{K}`-labelling `f` under birational
            rowmotion.

            EXAMPLES::

                sage: P = Poset({1: [2, 3], 2: [4], 3: [4]})
                sage: lex = [1, 2, 3, 4]
                sage: t = P.birational_free_labelling(linear_extension=lex); t
                (Fraction Field of Multivariate Polynomial Ring in a, x1, x2, x3, x4, b over Rational Field,
                 {...},
                 a,
                 b)
                sage: sorted(t[1].items())
                [(1, x1), (2, x2), (3, x3), (4, x4)]
                sage: t = P.birational_rowmotion(t); t
                (Fraction Field of Multivariate Polynomial Ring in a, x1, x2, x3, x4, b over Rational Field,
                 {...},
                 a,
                 b)
                sage: sorted(t[1].items())
                [(1, a*b/x4), (2, (x1*x2*b + x1*x3*b)/(x2*x4)),
                 (3, (x1*x2*b + x1*x3*b)/(x3*x4)), (4, (x2*b + x3*b)/x4)]

            A result of [GR2013]_ states that applying birational rowmotion
            `n+m` times to a `\mathbf{K}`-labelling `f` of the poset
            `[n] \times [m]` gives back `f`. Let us check this::

                sage: def test_rectangle_periodicity(n, m, k):
                ....:     P = posets.ChainPoset(n).product(posets.ChainPoset(m))
                ....:     t0 = P.birational_free_labelling(P)
                ....:     t = t0
                ....:     for i in range(k):
                ....:         t = P.birational_rowmotion(t)
                ....:     return t == t0
                sage: test_rectangle_periodicity(2, 2, 4)
                True
                sage: test_rectangle_periodicity(2, 2, 2)
                False
                sage: test_rectangle_periodicity(2, 3, 5)  # long time
                True

            While computations with the birational free labelling quickly
            run out of memory due to the complexity of the rational
            functions involved, it is computationally cheap to check
            properties of birational rowmotion on examples in the tropical
            semiring::

                sage: def test_rectangle_periodicity_tropical(n, m, k):
                ....:     P = posets.ChainPoset(n).product(posets.ChainPoset(m))
                ....:     TT = TropicalSemiring(ZZ)
                ....:     t0 = (TT, {v: TT(floor(random()*100)) for v in P}, TT(0), TT(124))
                ....:     t = t0
                ....:     for i in range(k):
                ....:         t = P.birational_rowmotion(t)
                ....:     return t == t0
                sage: test_rectangle_periodicity_tropical(7, 6, 13)
                True

            Tropicalization is also what relates birational rowmotion to
            classical rowmotion on order ideals. In fact, if `T` denotes
            the :class:`tropical semiring
            <sage.rings.semirings.tropical_semiring.TropicalSemiring>` of
            `\ZZ` and `P` is a finite poset, then we can define an embedding
            `\phi` from the set `J(P)` of all order ideals of `P` into the
            set `T^{\widehat{P}}` of all `T`-labellings of `P` by sending
            every `I \in J(P)` to the indicator function of `I` extended by
            the value `1` at the element `0` and the value `0` at the
            element `1`. This map `\phi` has the property that
            `R \circ \phi = \phi \circ r`, where `R` denotes birational
            rowmotion, and `r` denotes :meth:`classical rowmotion <rowmotion>`
            on `J(P)`. An example::

                sage: P = posets.IntegerPartitions(5)
                sage: TT = TropicalSemiring(ZZ)
                sage: def indicator_labelling(I):
                ....:     # send order ideal `I` to a `T`-labelling of `P`.
                ....:     dct = {v: TT(v in I) for v in P}
                ....:     return (TT, dct, TT(1), TT(0))
                sage: all(indicator_labelling(P.rowmotion(I))
                ....:     == P.birational_rowmotion(indicator_labelling(I))
                ....:     for I in P.order_ideals_lattice(facade=True))
                True

            TESTS:

            Facade set to false works::

                sage: P = Poset({1: [2, 3], 2: [4], 3: [4]}, facade=False)
                sage: lex = [1, 2, 3, 4]
                sage: t = P.birational_free_labelling(linear_extension=lex); t
                (Fraction Field of Multivariate Polynomial Ring in a, x1, x2, x3, x4, b over Rational Field,
                 {...},
                 a,
                 b)
                sage: t = P.birational_rowmotion(t); t
                (Fraction Field of Multivariate Polynomial Ring in a, x1, x2, x3, x4, b over Rational Field,
                 {...},
                 a,
                 b)
                sage: t[1][P(2)]
                (x1*x2*b + x1*x3*b)/(x2*x4)
                sage: t = P.birational_rowmotion(t)
                sage: t[1][P(2)]
                a*b/x3
            """
            l = labelling
            for v in reversed(self.linear_extension()):
                l = self.birational_toggle(v, l)
            return l

        def panyushev_orbits(self, element_constructor = set):
            r"""
            Return the Panyushev orbits of antichains in ``self``.

            The Panyushev orbit of an antichain is its orbit under
            Panyushev complementation (see
            :meth:`panyushev_complement`).

            INPUT:

            - ``element_constructor`` (defaults to ``set``) -- a type
              constructor (``set``, ``tuple``, ``list``, ``frozenset``,
              ``iter``, etc.) which is to be applied to the antichains
              before they are returned.

            OUTPUT:

            - the partition of the set of all antichains of ``self`` into
              orbits under Panyushev complementation. This is returned as
              a list of lists ``L`` such that for each ``L`` and ``i``,
              cyclically:
              ``self.order_ideal_complement_generators(L[i]) == L[i+1]``.
              The entries ``L[i]`` are sets by default, but depending on
              the optional keyword variable ``element_constructors``
              they can also be tuples, lists etc.

            EXAMPLES::

                sage: P = Poset( ( [1,2,3], [ [1,3], [2,3] ] ) )
                sage: orb = P.panyushev_orbits()
                sage: sorted(sorted(o) for o in orb)
                [[set(), {1, 2}, {3}], [{2}, {1}]]
                sage: orb = P.panyushev_orbits(element_constructor=list)
                sage: sorted(sorted(o) for o in orb)
                [[[], [1, 2], [3]], [[1], [2]]]
                sage: orb = P.panyushev_orbits(element_constructor=frozenset)
                sage: sorted(sorted(o) for o in orb)
                [[frozenset(), frozenset({1, 2}), frozenset({3})],
                 [frozenset({2}), frozenset({1})]]
                sage: orb = P.panyushev_orbits(element_constructor=tuple)
                sage: sorted(sorted(o) for o in orb)
                [[(), (1, 2), (3,)], [(1,), (2,)]]
                sage: P = Poset( {} )
                sage: P.panyushev_orbits()
                [[set()]]
            """
            # TODO: implement a generic function taking a set and
            # bijections on this set, and returning the orbits.
            AC = set(self.antichains(element_constructor = frozenset))
            orbits = []
            while AC:
                A = AC.pop()
                orbit = [ A ]
                while True:
                    A = frozenset(self.order_ideal_complement_generators(A))
                    if A not in AC:
                        break
                    orbit.append( A )
                    AC.remove( A )
                orbits.append([element_constructor(_) for _ in orbit])
            return orbits

        def rowmotion_orbits(self, element_constructor = set):
            r"""
            Return the rowmotion orbits of order ideals in ``self``.

            The rowmotion orbit of an order ideal is its orbit under
            rowmotion (see :meth:`rowmotion`).

            INPUT:

            - ``element_constructor`` (defaults to ``set``) -- a type
              constructor (``set``, ``tuple``, ``list``, ``frozenset``,
              ``iter``, etc.) which is to be applied to the antichains
              before they are returned.

            OUTPUT:

            - the partition of the set of all order ideals of ``self``
              into orbits under rowmotion. This is returned as
              a list of lists ``L`` such that for each ``L`` and ``i``,
              cyclically: ``self.rowmotion(L[i]) == L[i+1]``.
              The entries ``L[i]`` are sets by default, but depending on
              the optional keyword variable ``element_constructors``
              they can also be tuples, lists etc.

            EXAMPLES::

                sage: P = Poset( {1: [2, 3], 2: [], 3: [], 4: [2]} )
                sage: sorted(len(o) for o in P.rowmotion_orbits())
                [3, 5]
                sage: orb = P.rowmotion_orbits(element_constructor=list)
                sage: sorted(sorted(e) for e in orb)
                [[[], [4, 1], [4, 1, 2, 3]], [[1], [1, 3], [4], [4, 1, 2], [4, 1, 3]]]
                sage: orb = P.rowmotion_orbits(element_constructor=tuple)
                sage: sorted(sorted(e) for e in orb)
                [[(), (4, 1), (4, 1, 2, 3)], [(1,), (1, 3), (4,), (4, 1, 2), (4, 1, 3)]]
                sage: P = Poset({})
                sage: P.rowmotion_orbits(element_constructor=tuple)
                [[()]]
            """
            pan_orbits = self.panyushev_orbits(element_constructor = list)
            return [[element_constructor(self.order_ideal(oideal)) for oideal in orbit] for orbit in pan_orbits]

        def rowmotion_orbits_plots(self):
            r"""
            Return plots of the rowmotion orbits of order ideals in ``self``.

            The rowmotion orbit of an order ideal is its orbit under
            rowmotion (see :meth:`rowmotion`).

            EXAMPLES::

                sage: P = Poset( {1: [2, 3], 2: [], 3: [], 4: [2]} )
                sage: P.rowmotion_orbits_plots()
                Graphics Array of size 2 x 5
                sage: P = Poset({})
                sage: P.rowmotion_orbits_plots()
                Graphics Array of size 1 x 1

            """
            from sage.plot.plot import graphics_array
            plot_of_orb_plots=[]
            max_orbit_size = 0
            for orb in self.rowmotion_orbits():
                orb_plots=[]
                if len(orb) > max_orbit_size:
                    max_orbit_size = len(orb)
                for oi in orb:
                    oiplot = self.order_ideal_plot(oi)
                    orb_plots.append(oiplot)
                plot_of_orb_plots.append(orb_plots)
            return graphics_array(plot_of_orb_plots, ncols = max_orbit_size)


        def toggling_orbits(self, vs, element_constructor = set):
            r"""
            Return the orbits of order ideals in ``self`` under the
            operation of toggling the vertices ``vs[0], vs[1], ...``
            in this order.

            See :meth:`~sage.categories.posets.Posets.ParentMethods.order_ideal_toggle` for a definition of toggling.

            .. WARNING::

                The orbits are those under the composition of toggles,
                *not* under the single toggles themselves. Thus, for
                example, if ``vs == [1,2]``, then the orbits have the
                form `(I, T_2 T_1 I, T_2 T_1 T_2 T_1 I, \ldots)`
                (where `I` denotes an order ideal and `T_i` means
                toggling at `i`) rather than
                `(I, T_1 I, T_2 T_1 I, T_1 T_2 T_1 I, \ldots)`.

            INPUT:

            - ``vs``: a list (or other iterable) of elements of ``self``
              (but since the output depends on the order, sets should
              not be used as ``vs``).

            OUTPUT:

            - a partition of the order ideals of ``self``, as a list of
              sets ``L`` such that for each ``L`` and ``i``, cyclically:
              ``self.order_ideal_toggles(L[i], vs) == L[i+1]``.

            EXAMPLES::

                sage: P = Poset( {1: [2, 4], 2: [], 3: [4], 4: []} )
                sage: sorted(len(o) for o in P.toggling_orbits([1, 2]))
                [2, 3, 3]
                sage: P = Poset( {1: [3], 2: [1, 4], 3: [], 4: [3]} )
                sage: sorted(len(o) for o in P.toggling_orbits((1, 2, 4, 3)))
                [3, 3]
            """
            # TODO: implement a generic function taking a set and
            # bijections on this set, and returning the orbits.
            OI = set(self.order_ideals_lattice(facade=True))
            orbits = []
            while OI:
                A = OI.pop()
                orbit = [ A ]
                while True:
                    A = self.order_ideal_toggles(A, vs)
                    if A not in OI:
                        break
                    orbit.append( A )
                    OI.remove( A )
                orbits.append([element_constructor(_) for _ in orbit])
            return orbits

        def toggling_orbits_plots(self, vs):
            r"""
            Return plots of the orbits of order ideals in ``self`` under the
            operation of toggling the vertices ``vs[0], vs[1], ...``
            in this order.

            See :meth:`toggling_orbits` for more information.

            EXAMPLES::

                sage: P = Poset( {1: [2, 3], 2: [], 3: [], 4: [2]} )
                sage: P.toggling_orbits_plots([1,2,3,4])
                Graphics Array of size 2 x 5
                sage: P = Poset({})
                sage: P.toggling_orbits_plots([])
                Graphics Array of size 1 x 1

            """
            from sage.plot.plot import graphics_array
            plot_of_orb_plots=[]
            max_orbit_size = 0
            for orb in self.toggling_orbits(vs):
                orb_plots=[]
                if len(orb) > max_orbit_size:
                    max_orbit_size = len(orb)
                for oi in orb:
                    oiplot = self.order_ideal_plot(oi)
                    orb_plots.append(oiplot)
                plot_of_orb_plots.append(orb_plots)
            return graphics_array(plot_of_orb_plots, ncols = max_orbit_size)

        def panyushev_orbit_iter(self, antichain, element_constructor=set, stop=True, check=True):
            r"""
            Iterate over the Panyushev orbit of an antichain
            ``antichain`` of ``self``.

            The Panyushev orbit of an antichain is its orbit under
            Panyushev complementation (see
            :meth:`panyushev_complement`).

            INPUT:

            - ``antichain`` -- an antichain of ``self``, given as an
              iterable.

            - ``element_constructor`` (defaults to ``set``) -- a type
              constructor (``set``, ``tuple``, ``list``, ``frozenset``,
              ``iter``, etc.) which is to be applied to the antichains
              before they are yielded.

            - ``stop`` -- a Boolean (default: ``True``) determining
              whether the iterator should stop once it completes its
              cycle (this happens when it is set to ``True``) or go on
              forever (this happens when it is set to ``False``).

            - ``check`` -- a Boolean (default: ``True``) determining
              whether ``antichain`` should be checked for being an
              antichain.

            OUTPUT:

            - an iterator over the orbit of the antichain ``antichain``
              under Panyushev complementation. This iterator `I` has the
              property that ``I[0] == antichain`` and each `i` satisfies
              ``self.order_ideal_complement_generators(I[i]) == I[i+1]``,
              where ``I[i+1]`` has to be understood as ``I[0]`` if it is
              undefined.
              The entries ``I[i]`` are sets by default, but depending on
              the optional keyword variable ``element_constructors``
              they can also be tuples, lists etc.

            EXAMPLES::

                sage: P = Poset( ( [1,2,3], [ [1,3], [2,3] ] ) )
                sage: list(P.panyushev_orbit_iter(set([1, 2])))
                [{1, 2}, {3}, set()]
                sage: list(P.panyushev_orbit_iter([1, 2]))
                [{1, 2}, {3}, set()]
                sage: list(P.panyushev_orbit_iter([2, 1]))
                [{1, 2}, {3}, set()]
                sage: list(P.panyushev_orbit_iter(set([1, 2]), element_constructor=list))
                [[1, 2], [3], []]
                sage: list(P.panyushev_orbit_iter(set([1, 2]), element_constructor=frozenset))
                [frozenset({1, 2}), frozenset({3}), frozenset()]
                sage: list(P.panyushev_orbit_iter(set([1, 2]), element_constructor=tuple))
                [(1, 2), (3,), ()]

                sage: P = Poset( {} )
                sage: list(P.panyushev_orbit_iter([]))
                [set()]

                sage: P = Poset({ 1: [2, 3], 2: [4], 3: [4], 4: [] })
                sage: Piter = P.panyushev_orbit_iter([2], stop=False)
                sage: next(Piter)
                {2}
                sage: next(Piter)
                {3}
                sage: next(Piter)
                {2}
                sage: next(Piter)
                {3}
            """
            # TODO: implement a generic function taking a set and
            # bijections on this set, and returning an orbit of a given
            # element.
            if check:
                if not self.is_antichain_of_poset(antichain):
                    raise ValueError("the given antichain is not an antichain")
            starter = set(antichain)     # sanitize input
            yield element_constructor(starter)
            next = starter
            if stop:
                while True:
                    next = self.order_ideal_complement_generators(next)
                    if next == starter:
                        break
                    yield element_constructor(next)
            else:
                while True:
                    next = self.order_ideal_complement_generators(next)
                    yield element_constructor(next)

        def rowmotion_orbit_iter(self, oideal, element_constructor=set, stop=True, check=True):
            r"""
            Iterate over the rowmotion orbit of an order ideal
            ``oideal`` of ``self``.

            The rowmotion orbit of an order ideal is its orbit under
            rowmotion (see :meth:`rowmotion`).

            INPUT:

            - ``oideal`` -- an order ideal of ``self``, given as an
              iterable.

            - ``element_constructor`` (defaults to ``set``) -- a type
              constructor (``set``, ``tuple``, ``list``, ``frozenset``,
              ``iter``, etc.) which is to be applied to the order
              ideals before they are yielded.

            - ``stop`` -- a Boolean (default: ``True``) determining
              whether the iterator should stop once it completes its
              cycle (this happens when it is set to ``True``) or go on
              forever (this happens when it is set to ``False``).

            - ``check`` -- a Boolean (default: ``True``) determining
              whether ``oideal`` should be checked for being an
              order ideal.

            OUTPUT:

            - an iterator over the orbit of the order ideal ``oideal``
              under rowmotion. This iterator `I` has the property that
              ``I[0] == oideal`` and that every `i` satisfies
              ``self.rowmotion(I[i]) == I[i+1]``, where ``I[i+1]`` has
              to be understood as ``I[0]`` if it is undefined.
              The entries ``I[i]`` are sets by default, but depending on
              the optional keyword variable ``element_constructors``
              they can also be tuples, lists etc.

            EXAMPLES::

                sage: P = Poset( ( [1,2,3], [ [1,3], [2,3] ] ) )
                sage: list(P.rowmotion_orbit_iter(set([1, 2])))
                [{1, 2}, {1, 2, 3}, set()]
                sage: list(P.rowmotion_orbit_iter([1, 2]))
                [{1, 2}, {1, 2, 3}, set()]
                sage: list(P.rowmotion_orbit_iter([2, 1]))
                [{1, 2}, {1, 2, 3}, set()]
                sage: list(P.rowmotion_orbit_iter(set([1, 2]), element_constructor=list))
                [[1, 2], [1, 2, 3], []]
                sage: list(P.rowmotion_orbit_iter(set([1, 2]), element_constructor=frozenset))
                [frozenset({1, 2}), frozenset({1, 2, 3}), frozenset()]
                sage: list(P.rowmotion_orbit_iter(set([1, 2]), element_constructor=tuple))
                [(1, 2), (1, 2, 3), ()]

                sage: P = Poset( {} )
                sage: list(P.rowmotion_orbit_iter([]))
                [set()]

                sage: P = Poset({ 1: [2, 3], 2: [4], 3: [4], 4: [] })
                sage: Piter = P.rowmotion_orbit_iter([1, 2, 3], stop=False)
                sage: next(Piter)
                {1, 2, 3}
                sage: next(Piter)
                {1, 2, 3, 4}
                sage: next(Piter)
                set()
                sage: next(Piter)
                {1}
                sage: next(Piter)
                {1, 2, 3}

                sage: P = Poset({ 1: [4], 2: [4, 5], 3: [5] })
                sage: list(P.rowmotion_orbit_iter([1, 2], element_constructor=list))
                [[1, 2], [1, 2, 3, 4], [2, 3, 5], [1], [2, 3], [1, 2, 3, 5], [1, 2, 4], [3]]
            """
            # TODO: implement a generic function taking a set and
            # bijections on this set, and returning an orbit of a given
            # element.
            if check:
                if not self.is_order_ideal(oideal):
                    raise ValueError("the given order ideal is not an order ideal")
            starter = set(oideal)     # sanitize input
            yield element_constructor(starter)
            next = starter
            if stop:
                while True:
                    next = self.rowmotion(next)
                    if next == starter:
                        break
                    yield element_constructor(next)
            else:
                while True:
                    next = self.rowmotion(next)
                    yield element_constructor(next)

        def toggling_orbit_iter(self, vs, oideal, element_constructor=set, stop=True, check=True):
            r"""
            Iterate over the orbit of an order ideal ``oideal`` of
            ``self`` under the operation of toggling the vertices
            ``vs[0], vs[1], ...`` in this order.

            See :meth:`~sage.categories.posets.Posets.ParentMethods.order_ideal_toggle` for a definition of toggling.

            .. WARNING::

                The orbit is that under the composition of toggles,
                *not* under the single toggles themselves. Thus, for
                example, if ``vs == [1,2]``, then the orbit has the
                form `(I, T_2 T_1 I, T_2 T_1 T_2 T_1 I, \ldots)`
                (where `I` denotes ``oideal`` and `T_i` means
                toggling at `i`) rather than
                `(I, T_1 I, T_2 T_1 I, T_1 T_2 T_1 I, \ldots)`.

            INPUT:

            - ``vs``: a list (or other iterable) of elements of ``self``
              (but since the output depends on the order, sets should
              not be used as ``vs``).

            - ``oideal`` -- an order ideal of ``self``, given as an
              iterable.

            - ``element_constructor`` (defaults to ``set``) -- a type
              constructor (``set``, ``tuple``, ``list``, ``frozenset``,
              ``iter``, etc.) which is to be applied to the order
              ideals before they are yielded.

            - ``stop`` -- a Boolean (default: ``True``) determining
              whether the iterator should stop once it completes its
              cycle (this happens when it is set to ``True``) or go on
              forever (this happens when it is set to ``False``).

            - ``check`` -- a Boolean (default: ``True``) determining
              whether ``oideal`` should be checked for being an
              order ideal.

            OUTPUT:

            - an iterator over the orbit of the order ideal ``oideal``
              under toggling the vertices in the list ``vs`` in this
              order. This iterator `I` has the property that
              ``I[0] == oideal`` and that every `i` satisfies
              ``self.order_ideal_toggles(I[i], vs) == I[i+1]``, where
              ``I[i+1]`` has to be understood as ``I[0]`` if it is
              undefined.
              The entries ``I[i]`` are sets by default, but depending on
              the optional keyword variable ``element_constructors``
              they can also be tuples, lists etc.

            EXAMPLES::

                sage: P = Poset( ( [1,2,3], [ [1,3], [2,3] ] ) )
                sage: list(P.toggling_orbit_iter([1, 3, 1], set([1, 2])))
                [{1, 2}]
                sage: list(P.toggling_orbit_iter([1, 2, 3], set([1, 2])))
                [{1, 2}, set(), {1, 2, 3}]
                sage: list(P.toggling_orbit_iter([3, 2, 1], set([1, 2])))
                [{1, 2}, {1, 2, 3}, set()]
                sage: list(P.toggling_orbit_iter([3, 2, 1], set([1, 2]), element_constructor=list))
                [[1, 2], [1, 2, 3], []]
                sage: list(P.toggling_orbit_iter([3, 2, 1], set([1, 2]), element_constructor=frozenset))
                [frozenset({1, 2}), frozenset({1, 2, 3}), frozenset()]
                sage: list(P.toggling_orbit_iter([3, 2, 1], set([1, 2]), element_constructor=tuple))
                [(1, 2), (1, 2, 3), ()]
                sage: list(P.toggling_orbit_iter([3, 2, 1], [2, 1], element_constructor=tuple))
                [(1, 2), (1, 2, 3), ()]

                sage: P = Poset( {} )
                sage: list(P.toggling_orbit_iter([], []))
                [set()]

                sage: P = Poset({ 1: [2, 3], 2: [4], 3: [4], 4: [] })
                sage: Piter = P.toggling_orbit_iter([1, 2, 4, 3], [1, 2, 3], stop=False)
                sage: next(Piter)
                {1, 2, 3}
                sage: next(Piter)
                {1}
                sage: next(Piter)
                set()
                sage: next(Piter)
                {1, 2, 3}
                sage: next(Piter)
                {1}
            """
            # TODO: implement a generic function taking a set and
            # bijections on this set, and returning an orbit of a given
            # element.
            if check:
                if not self.is_order_ideal(oideal):
                    raise ValueError("the given order ideal is not an order ideal")
            starter = set(oideal)     # sanitize input
            yield element_constructor(starter)
            next = starter
            if stop:
                while True:
                    next = self.order_ideal_toggles(next, vs)
                    if next == starter:
                        break
                    yield element_constructor(next)
            else:
                while True:
                    next = self.order_ideal_toggles(next, vs)
                    yield element_constructor(next)

        def order_ideals_lattice(self, as_ideals=True, facade=None):
            r"""
            Return the lattice of order ideals of a poset ``self``,
            ordered by inclusion.

            The lattice of order ideals of a poset `P` is usually
            denoted by `J(P)`. Its underlying set is the set of order
            ideals of `P`, and its partial order is given by
            inclusion.

            The order ideals of `P` are in a canonical bijection
            with the antichains of `P`. The bijection maps every
            order ideal to the antichain formed by its maximal
            elements. By setting the ``as_ideals`` keyword variable to
            ``False``, one can make this method apply this bijection
            before returning the lattice.

            INPUT:

            - ``as_ideals`` -- Boolean, if ``True`` (default) returns
              a poset on the set of order ideals, otherwise on the set
              of antichains
            - ``facade`` -- Boolean or ``None`` (default). Whether to
              return a facade lattice or not. By default return facade
              lattice if the poset is a facade poset.

            EXAMPLES::

                sage: P = posets.PentagonPoset()
                sage: P.cover_relations()
                [[0, 1], [0, 2], [1, 4], [2, 3], [3, 4]]
                sage: J = P.order_ideals_lattice(); J
                Finite lattice containing 8 elements
                sage: sorted(sorted(e) for e in J)
                 [[], [0], [0, 1], [0, 1, 2], [0, 1, 2, 3], [0, 1, 2, 3, 4], [0, 2], [0, 2, 3]]

            As a lattice on antichains::

                sage: J2 = P.order_ideals_lattice(False); J2
                Finite lattice containing 8 elements
                sage: sorted(J2)
                [(), (0,), (1,), (1, 2), (1, 3), (2,), (3,), (4,)]

            TESTS::

                sage: J = posets.DiamondPoset(4, facade = True).order_ideals_lattice(); J
                Finite lattice containing 6 elements
                sage: sorted(sorted(e) for e in J)
                [[], [0], [0, 1], [0, 1, 2], [0, 1, 2, 3], [0, 2]]
                sage: sorted(sorted(sorted(e) for e in c) for c in J.cover_relations())
                [[[], [0]], [[0], [0, 1]], [[0], [0, 2]], [[0, 1], [0, 1, 2]], [[0, 1, 2], [0, 1, 2, 3]], [[0, 1, 2], [0, 2]]]

                sage: P = Poset({1:[2]})
                sage: J_facade = P.order_ideals_lattice()
                sage: J_nonfacade = P.order_ideals_lattice(facade=False)
                sage: type(J_facade[0]) == type(J_nonfacade[0])
                False
            """
            from sage.combinat.posets.lattices import LatticePoset
            if facade is None:
                facade = self._is_facade
            if as_ideals:
                from sage.misc.call import attrcall
                from sage.sets.set import Set
                ideals = [Set(self.order_ideal(antichain))
                          for antichain in self.antichains()]
                return LatticePoset((ideals, attrcall("issubset")),
                                    facade=facade)
            else:
                from sage.misc.cachefunc import cached_function
                antichains = [tuple(a) for a in self.antichains()]
                @cached_function
                def is_above(a, xb):
                    return any(self.is_lequal(xa, xb) for xa in a)
                def compare(a, b):
                    return all(is_above(a, xb) for xb in b)
                return LatticePoset((antichains, compare), facade=facade)

        @abstract_method(optional = True)
        def antichains(self):
            r"""
            Return all antichains of ``self``.

            EXAMPLES::

                sage: A = posets.PentagonPoset().antichains(); A
                Set of antichains of Finite lattice containing 5 elements
                sage: list(A)
                [[], [0], [1], [1, 2], [1, 3], [2], [3], [4]]
            """

        def directed_subsets(self, direction):
            r"""
            Return the order filters (resp. order ideals) of ``self``, as lists.

            If ``direction`` is 'up', returns the order filters (upper sets).

            If ``direction`` is 'down', returns the order ideals (lower sets).

            INPUT:

            - ``direction`` -- 'up' or 'down'

            EXAMPLES::

                sage: P = Poset((divisors(12), attrcall("divides")), facade=True)
                sage: A = P.directed_subsets('up')
                sage: sorted(list(A))
                [[], [1, 2, 4, 3, 6, 12], [2, 4, 3, 6, 12], [2, 4, 6, 12], [3, 6, 12], [4, 3, 6, 12], [4, 6, 12], [4, 12], [6, 12], [12]]

            TESTS::

                sage: list(Poset().directed_subsets('up'))
                [[]]
            """
            if direction != 'up' and direction != 'down':
                raise ValueError("Direction must be either 'up' or 'down'.")
            return self.antichains().map(lambda elements: self.directed_subset(elements, direction))
